from django.urls import path
from django.contrib.auth import views as auth_views
from django.conf import settings

from . import views
from .app_views import urlpatterns as app_urlpatterns

# Standardwerte laden
try:
    from .models import Einstellung, Geheime_Einstellung
    Einstellung.objects.get_or_create(id="wc-url", typ="url", name="WooCommerce Shop-Url")[0]

    Geheime_Einstellung.objects.get_or_create(id="wc-consumer_key")
    Geheime_Einstellung.objects.get_or_create(id="wc-consumer_secret")
    Geheime_Einstellung.objects.get_or_create(id="wc-url")
except:
    pass


#######################

app_name = 'kmuhelper'
urlpatterns = [
    path('wc/auth/start', views.wc_auth_start, name="wc-auth-start"),
    path('wc/auth/end', views.wc_auth_end, name="wc-auth-end"),
    path('wc/auth/key', views.wc_auth_key, name="wc-auth-key"),
    path('wc/import/products', views.wc_import_products, name="wc-import-products"),
    path('wc/import/customers', views.wc_import_customers, name="wc-import-customers"),
    path('wc/import/categories', views.wc_import_categories, name="wc-import-categories"),
    path('wc/import/orders', views.wc_import_orders, name="wc-import-orders"),
    path('wc/update/product/<object_id>', views.wc_update_product, name="wc-update-product"),
    path('wc/update/customer/<object_id>', views.wc_update_customer, name="wc-update-customer"),
    path('wc/update/category/<object_id>', views.wc_update_category, name="wc-update-category"),
    path('wc/update/order/<object_id>', views.wc_update_order, name="wc-update-order"),
    path('wc/webhooks', views.wc_webhooks, name="wc-webhooks"),
    path('kunde/<object_id>/email-registriert', views.kunde_email_registriert, name='kunde-email-registriert'),
    path('lieferant/<object_id>/zuordnen', views.lieferant_zuordnen, name='lieferant-zuordnen'),
    path('lieferung/<object_id>/einlagern', views.lieferung_einlagern, name='lieferung-einlagern'),
    path('bestellung/<object_id>/pdf', views.bestellung_pdf_ansehen, name='bestellung-pdf-ansehen'),
    path('bestellung/<object_id>/pdf/ankundensenden', views.bestellung_pdf_an_kunden_senden, name='bestellung-pdf-an-kunden-senden'),

    path('kunde/bestellung/<order_id>/<order_key>/', views.kunde_rechnung_ansehen, name='kunde-rechnung-ansehen'),
] + app_urlpatterns



from .models import ToDoNotiz, ToDoVersand, ToDoZahlungseingang, ToDoLagerbestand, ToDoLieferung
from django.contrib import admin, messages
from datetime import datetime
from pytz import utc

from .models import Ansprechpartner, Bestellung, Kategorie, Kosten, Kunde, Lieferant, Lieferung, Notiz, Produkt, Zahlungsempfaenger, Einstellung
from .apis import WooCommerce

# Disable "view on site" globally

admin.site.site_url = None

# Register your models here.


@admin.register(Ansprechpartner)
class AnsprechpartnerAdmin(admin.ModelAdmin):
    fieldsets = [
        ("Name", {'fields': ['name']}),
        ('Daten', {'fields': ['telefon', 'email']})
    ]

    ordering = ('name',)

    list_display = ('name', 'telefon', 'email')
    search_fields = ['name', 'telefon', 'email']


class BestellungInlineBestellungsposten(admin.TabularInline):
    model = Bestellung.produkte.through
    verbose_name = "Bestellungsposten"
    verbose_name_plural = "Bestellungsposten"
    extra = 0

    readonly_fields = ["zwischensumme", "mwstsatz", "produkt", "produktpreis"]
    fieldsets = [(None, {'fields': ['produkt', 'bemerkung',
                                    'produktpreis', 'menge', 'mwstsatz', 'zwischensumme']})]

    def has_change_permission(self, request, obj=None):
        return False if (obj and (obj.versendet or obj.bezahlt)) else True

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False if (obj and (obj.versendet or obj.bezahlt)) else True


class BestellungInlineBestellungspostenAdd(admin.TabularInline):
    model = Bestellung.produkte.through
    verbose_name = "Bestellungsposten"
    verbose_name_plural = "Bestellungsposten hinzufügen"
    extra = 0

    raw_id_fields = ("produkt",)

    fieldsets = [(None, {'fields': ['produkt', 'bemerkung', 'menge']})]

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False if (obj and (obj.versendet or obj.bezahlt)) else True

    def has_delete_permission(self, request, obj=None):
        return False

    def has_view_permission(self, request, obj=None):
        return False


class BestellungInlineBestellungskosten(admin.TabularInline):
    model = Bestellung.kosten.through
    verbose_name = "Bestellungskosten"
    verbose_name_plural = "Bestellungskosten"
    extra = 0

    readonly_fields = ["zwischensumme", "mwstsatz", "kosten"]
    fieldsets = [
        (None, {'fields': ['kosten', 'bemerkung', 'mwstsatz', 'zwischensumme']})]

    def has_change_permission(self, request, obj=None):
        return False if (obj and (obj.versendet or obj.bezahlt)) else True

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False if (obj and (obj.versendet or obj.bezahlt)) else True


class BestellungInlineBestellungskostenAdd(admin.TabularInline):
    model = Bestellung.kosten.through
    verbose_name = "Bestellungskosten"
    verbose_name_plural = "Bestellungskosten hinzufügen"
    extra = 0

    fieldsets = [(None, {'fields': ['kosten', 'bemerkung']})]

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False if (obj and (obj.versendet or obj.bezahlt)) else True

    def has_delete_permission(self, request, obj=None):
        return False

    def has_view_permission(self, request, obj=None):
        return False


@admin.register(Bestellung)
class BestellungsAdmin(admin.ModelAdmin):
    list_display = ('id', 'datum', 'kunde', 'status', 'zahlungsmethode',
                    'versendet', 'bezahlt', 'fix_summe', 'html_notiz')
    list_filter = ('status', 'bezahlt', 'versendet', 'zahlungsmethode')
    search_fields = ['id', 'datum', 'notiz__name', 'notiz__beschrieb', 'kundennotiz', 'trackingnummer', 'rechnungsadresse_vorname',
                     'rechnungsadresse_nachname', 'rechnungsadresse_firma', 'rechnungsadresse_plz', 'rechnungsadresse_ort']

    ordering = ("versendet", "bezahlt", "-datum")

    inlines = [BestellungInlineBestellungsposten, BestellungInlineBestellungspostenAdd,
               BestellungInlineBestellungskosten, BestellungInlineBestellungskostenAdd]

    raw_id_fields = ("kunde",)

    save_on_top = True

    list_select_related = ["kunde"]

    def get_fieldsets(self, request, obj=None):
        if obj:
            return [
                ('Einstellungen', {'fields': [
                 'zahlungsempfaenger', 'ansprechpartner']}),
                ('Infos', {'fields': ['name', 'datum', 'status']}),
                ('Lieferung', {'fields': ['versendet', 'trackingnummer']}),
                ('Bezahlung', {'fields': [
                 'bezahlt', 'zahlungsmethode', 'summe', 'summe_mwst', 'summe_gesamt']}),
                ('Kunde', {'fields': ['kunde']}),
                ('Notizen', {'fields': ['kundennotiz', 'html_notiz'], 'classes': [
                 "collapse start-open"]}),
                ('Rechnungsadresse', {'fields': ['rechnungsadresse_vorname', 'rechnungsadresse_nachname', 'rechnungsadresse_firma', 'rechnungsadresse_adresszeile1', 'rechnungsadresse_adresszeile2',
                                                 'rechnungsadresse_plz', 'rechnungsadresse_ort', 'rechnungsadresse_kanton', 'rechnungsadresse_land', 'rechnungsadresse_email', 'rechnungsadresse_telefon'], 'classes': ["collapse default-open"]}),
                ('Lieferadresse', {'fields': ['lieferadresse_vorname', 'lieferadresse_nachname', 'lieferadresse_firma', 'lieferadresse_adresszeile1',
                                              'lieferadresse_adresszeile2', 'lieferadresse_plz', 'lieferadresse_ort', 'lieferadresse_kanton', 'lieferadresse_land'], 'classes': ["collapse start-open"]})
            ]
        else:
            return [
                ('Einstellungen', {'fields': [
                 'zahlungsempfaenger', 'ansprechpartner']}),
                ('Lieferung', {'fields': ['trackingnummer']}),
                ('Bezahlung', {'fields': ['bezahlt', 'zahlungsmethode']}),
                ('Kunde', {'fields': ['kunde']}),
                ('Notizen', {'fields': ['kundennotiz'],
                             'classes': ["collapse start-open"]}),
            ]

    def get_readonly_fields(self, request, obj=None):
        rechnungsadresse = ['rechnungsadresse_vorname', 'rechnungsadresse_nachname', 'rechnungsadresse_firma', 'rechnungsadresse_adresszeile1', 'rechnungsadresse_adresszeile2',
                            'rechnungsadresse_plz', 'rechnungsadresse_ort', 'rechnungsadresse_kanton', 'rechnungsadresse_land', 'rechnungsadresse_email', 'rechnungsadresse_telefon']
        lieferadresse = ['lieferadresse_vorname', 'lieferadresse_nachname', 'lieferadresse_firma', 'lieferadresse_adresszeile1',
                         'lieferadresse_adresszeile2', 'lieferadresse_plz', 'lieferadresse_ort', 'lieferadresse_kanton', 'lieferadresse_land']
        fields = ['html_notiz', 'name', 'trackinglink',
                  'summe', 'summe_mwst', 'summe_gesamt']
        if obj:
            if obj.versendet:
                fields += ['versendet', 'trackingnummer'] + lieferadresse
            if obj.bezahlt:
                fields += ['bezahlt', 'zahlungsmethode'] + rechnungsadresse
            if obj.woocommerceid:
                fields += ["kundennotiz"]
        else:
            fields += ["status"]
        return fields

    def als_bezahlt_markieren(self, request, queryset):
        successcount = 0
        errorcount = 0
        for bestellung in queryset.all():
            if bestellung.bezahlt:
                errorcount += 1
            else:
                bestellung.bezahlt = True
                bestellung.save()
                successcount += 1
        messages.success(request, (('{} Bestellungen' if successcount !=
                                    1 else 'Eine Bestellung') + ' als bezahlt markiert!').format(successcount))
        if errorcount:
            messages.error(request, ('Beim als bezahlt markieren von ' + ('{} Bestellungen' if errorcount !=
                                                                          1 else 'einer Bestellung') + ' ist ein Fehler aufgetreten!').format(errorcount))
    als_bezahlt_markieren.short_description = "Als bezahlt markieren"

    def wc_update(self, request, queryset):
        result = WooCommerce.order_bulk_update(queryset.all())
        messages.success(request, ((
            '{} Bestellungen' if result[0] != 1 else 'Eine Bestellung') + ' von WooCommerce aktualisiert!').format(result[0]))
        if result[1]:
            messages.error(request, ('Beim Import von ' + (
                '{} Bestellungen' if result[1] != 1 else 'einer Bestellung') + ' von WooCommerce ist ein Fehler aufgetreten!').format(result[1]))
    wc_update.short_description = "Bestellungen von WooCommerce aktualisieren"

    actions = [als_bezahlt_markieren, wc_update]


class KategorienInlineUntergeordneteKategorien(admin.TabularInline):
    model = Produkt.kategorien.through
    verbose_name = "Produkt in dieser Kategorie"
    verbose_name_plural = "Produkte in dieser Kategorie"
    extra = 0


@admin.register(Kategorie)
class KategorienAdmin(admin.ModelAdmin):
    fieldsets = [
        ('Infos', {'fields': ['name', 'beschrieb', 'bildlink']}),
        ('Übergeordnete Kategorie', {'fields': ['uebergeordnete_kategorie']})
    ]

    list_display = ('name', 'beschrieb',
                    'uebergeordnete_kategorie', 'bild', 'anzahl_produkte')
    search_fields = ['name', 'beschrieb']

    ordering = ("uebergeordnete_kategorie", "name")

    inlines = [KategorienInlineUntergeordneteKategorien]

    list_select_related = ("uebergeordnete_kategorie",)

    def wc_update(self, request, queryset):
        result = WooCommerce.category_bulk_update(queryset.all())
        messages.success(request, ((
            '{} Kategorien' if result[0] != 1 else 'Eine Kategorie') + ' von WooCommerce aktualisiert!').format(result[0]))
        if result[1]:
            messages.error(request, ('Beim Import von ' + (
                '{} Kategorien' if result[1] != 1 else 'einer Kategorie') + ' von WooCommerce ist ein Fehler aufgetreten!').format(result[1]))
    wc_update.short_description = "Kategorien von WooCommerce aktualisieren"

    actions = ["wc_update"]


@admin.register(Kosten)
class KostenAdmin(admin.ModelAdmin):
    fieldsets = [
        (None, {"fields": ("name", "preis", "mwstsatz")})
    ]

    def get_readonly_fields(self, request, obj=None):
        return ("preis", "versteuerbar") if obj else []


@admin.register(Kunde)
class KundenAdmin(admin.ModelAdmin):
    def get_fieldsets(self, request, obj=None):
        if obj:
            return [
                ('Infos', {'fields': ['vorname', 'nachname',
                                      'firma', 'email', 'benutzername', 'sprache']}),
                ('Rechnungsadresse', {'fields': [('rechnungsadresse_vorname', 'rechnungsadresse_nachname'), 'rechnungsadresse_firma', ('rechnungsadresse_adresszeile1', 'rechnungsadresse_adresszeile2'), (
                    'rechnungsadresse_plz', 'rechnungsadresse_ort'), ('rechnungsadresse_kanton', 'rechnungsadresse_land'), ('rechnungsadresse_email', 'rechnungsadresse_telefon')]}),
                ('Lieferadresse', {'fields': [('lieferadresse_vorname', 'lieferadresse_nachname'), 'lieferadresse_firma', ('lieferadresse_adresszeile1', 'lieferadresse_adresszeile2'), (
                    'lieferadresse_plz', 'lieferadresse_ort'), ('lieferadresse_kanton', 'lieferadresse_land')], 'classes': ["collapse start-open"]}),
                ('Diverses', {'fields': [
                 'webseite', 'bemerkung', 'html_notiz']}),
                ('Erweitert', {'fields': [
                 'zusammenfuegen'], 'classes': ["collapse"]})
            ]
        else:
            return [
                ('Infos', {'fields': ['vorname', 'nachname',
                                      'firma', 'email', 'benutzername', 'sprache']}),
                ('Rechnungsadresse', {'fields': [('rechnungsadresse_vorname', 'rechnungsadresse_nachname'), 'rechnungsadresse_firma', ('rechnungsadresse_adresszeile1', 'rechnungsadresse_adresszeile2'), (
                    'rechnungsadresse_plz', 'rechnungsadresse_ort'), ('rechnungsadresse_kanton', 'rechnungsadresse_land'), ('rechnungsadresse_email', 'rechnungsadresse_telefon')]}),
                ('Lieferadresse', {'fields': [('lieferadresse_vorname', 'lieferadresse_nachname'), 'lieferadresse_firma', ('lieferadresse_adresszeile1', 'lieferadresse_adresszeile2'), (
                    'lieferadresse_plz', 'lieferadresse_ort'), ('lieferadresse_kanton', 'lieferadresse_land')], 'classes': ["collapse start-open"]}),
                ('Diverses', {'fields': ['webseite', 'bemerkung']})
            ]

    ordering = ('rechnungsadresse_plz', 'firma', 'nachname', 'vorname')

    list_display = ('id', 'firma', 'nachname', 'vorname', 'rechnungsadresse_plz',
                    'rechnungsadresse_ort', 'email', 'avatar', 'html_notiz')
    search_fields = ['id', 'nachname', 'vorname', 'firma', 'email', 'benutzername', 'rechnungsadresse_vorname', 'rechnungsadresse_nachname', 'rechnungsadresse_firma', 'rechnungsadresse_adresszeile1', 'rechnungsadresse_adresszeile2', 'rechnungsadresse_ort', 'rechnungsadresse_kanton', 'rechnungsadresse_plz', 'rechnungsadresse_land',
                     'rechnungsadresse_email', 'rechnungsadresse_telefon', 'lieferadresse_vorname', 'lieferadresse_nachname', 'lieferadresse_firma', 'lieferadresse_adresszeile1', 'lieferadresse_adresszeile2', 'lieferadresse_ort', 'lieferadresse_kanton', 'lieferadresse_kanton', 'lieferadresse_plz', 'lieferadresse_land', 'webseite', 'notiz__name', 'notiz__beschrieb']

    readonly_fields = ["html_notiz"]

    actions = ["wc_update"]

    save_on_top = True

    def wc_update(self, request, queryset):
        result = WooCommerce.customer_bulk_update(queryset.all())
        messages.success(request, ((
            '{} Kunden' if result[0] != 1 else 'Ein Kunde') + ' von WooCommerce aktualisiert!').format(result[0]))
        if result[1]:
            messages.error(request, ('Beim Import von ' + (
                '{} Kunden' if result[1] != 1 else 'einem Kunden') + ' von WooCommerce ist ein Fehler aufgetreten!').format(result[1]))
    wc_update.short_description = "Kunden von WooCommerce aktualisieren"


@admin.register(Lieferant)
class LieferantenAdmin(admin.ModelAdmin):
    fieldsets = [
        ('Infos', {'fields': ['kuerzel', 'name']}),
        ('Firma', {'fields': ['webseite', 'telefon', 'email']}),
        ('Texte', {'fields': ['adresse', 'notiz'], 'classes': ["collapse"]}),
        ('Ansprechpartner', {'fields': [
         'ansprechpartner', 'ansprechpartnertel', 'ansprechpartnermail'], 'classes': ["collapse"]})
    ]

    ordering = ('kuerzel',)

    list_display = ('kuerzel', 'name', 'notiz')
    search_fields = ['kuerzel', 'name', 'adresse', 'notiz']


class LieferungInlineProdukte(admin.TabularInline):
    model = Lieferung.produkte.through
    verbose_name = "Produkt"
    verbose_name_plural = "Produkte"
    extra = 0

    raw_id_fields = ("produkt",)

    def has_change_permission(self, request, obj=None):
        if obj and obj.eingelagert:
            return False
        else:
            return True

    def has_add_permission(self, request, obj=None):
        if obj and obj.eingelagert:
            return False
        else:
            return True

    def has_delete_permission(self, request, obj=None):
        if obj and obj.eingelagert:
            return False
        else:
            return True


@admin.register(Lieferung)
class LieferungenAdmin(admin.ModelAdmin):
    list_display = ('name', 'datum', 'notiz', 'anzahlprodukte', 'eingelagert')
    list_filter = ("eingelagert",)

    search_fields = ["name", "datum", "notiz"]

    ordering = ('name',)

    fieldsets = [
        ('Infos', {'fields': ['name', 'notiz']}),
        ('Lieferant', {'fields': ['lieferant']})
    ]

    inlines = [LieferungInlineProdukte]

    save_on_top = True

    actions = ["einlagern"]

    def einlagern(self, request, queryset):
        successcount = 0
        errorcount = 0
        for lieferung in queryset.all():
            if lieferung.einlagern():
                successcount += 1
            else:
                errorcount += 1
        messages.success(request, (('{} Lieferungen' if successcount !=
                                    1 else 'Eine Lieferung') + ' wurde eingelagert!').format(successcount))
        if errorcount:
            messages.error(request, ('Beim Einlagern von ' + ('{} Lieferungen' if errorcount !=
                                                              1 else 'einer Lieferung') + ' ist ein Fehler aufgetreten!').format(errorcount))
    einlagern.short_description = "Lieferungen einlagern"


@admin.register(Notiz)
class NotizenAdmin(admin.ModelAdmin):
    list_display = ["name", "beschrieb", "priority", "erledigt", "erstellt_am"]
    list_filter = ["erledigt", "priority"]

    readonly_fields = ["links"]

    ordering = ["erledigt", "-priority", "erstellt_am"]

    search_fields = ("name", "beschrieb")

    def get_fieldsets(self, request, obj=None):
        if obj:
            return [
                ("Infos", {"fields": ["name", "beschrieb"]}),
                ("Daten", {"fields": ["erledigt", "priority"]}),
                ("Verknüpfungen", {"fields": ["links"]}),
            ]
        else:
            return [
                ("Infos", {"fields": ["name", "beschrieb"]}),
                ("Daten", {"fields": ["erledigt", "priority"]}),
            ]

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if obj is None:
            if "from_bestellung" in request.GET:
                form.base_fields['name'].initial = 'Bestellung #' + \
                    request.GET.get("from_bestellung")
                form.base_fields['beschrieb'].initial = '\n\nDiese Notiz gehört zu Bestellung #' + \
                    request.GET.get("from_bestellung")
            if "from_produkt" in request.GET:
                form.base_fields['name'].initial = 'Produkt #' + \
                    request.GET.get("from_produkt")
                form.base_fields['beschrieb'].initial = '\n\nDiese Notiz gehört zu Produkt ' + \
                    request.GET.get("from_produkt")
            if "from_kunde" in request.GET:
                form.base_fields['name'].initial = 'Kunde #' + \
                    request.GET.get("from_kunde")
                form.base_fields['beschrieb'].initial = '\n\nDiese Notiz gehört zu Kunde ' + \
                    request.GET.get("from_kunde")
        return form

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not change:
            if "from_bestellung" in request.GET:
                if Bestellung.objects.filter(pk=request.GET["from_bestellung"]).exists():
                    bestellung = Bestellung.objects.get(
                        pk=request.GET["from_bestellung"])
                    obj.bestellung = bestellung
                    obj.save()
                    messages.info(
                        request, "Bestellung #" + str(bestellung.pk) + " wurde mit dieser Notiz verknüpft.")
                else:
                    messages.warning(
                        request, "Bestellung #" + request.GET["from_bestellung"] + " konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt.")
            if "from_produkt" in request.GET:
                if Produkt.objects.filter(pk=request.GET["from_produkt"]).exists():
                    produkt = Produkt.objects.get(
                        pk=request.GET["from_produkt"])
                    obj.produkt = produkt
                    obj.save()
                    messages.info(
                        request, "Produkt #" + str(produkt.pk) + " wurde mit dieser Notiz verknüpft.")
                else:
                    messages.warning(
                        request, "Produkt #" + request.GET["from_produkt"] + " konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt.")
            if "from_kunde" in request.GET:
                if Kunde.objects.filter(pk=request.GET["from_kunde"]).exists():
                    kunde = Kunde.objects.get(pk=request.GET["from_kunde"])
                    obj.kunde = kunde
                    obj.save()
                    messages.info(request, "Kunde #" + str(kunde.pk) +
                                  " wurde mit dieser Notiz verknüpft.")
                else:
                    messages.warning(
                        request, "Kunde #" + request.GET["from_kunde"] + " konnte nicht gefunden werden. Die Notiz wurde trotzdem erstellt.")


class ProduktInlineProduktkategorien(admin.TabularInline):
    model = Produkt.kategorien.through
    extra = 0


@admin.register(Produkt)
class ProduktAdmin(admin.ModelAdmin):
    def get_fieldsets(self, request, obj=None):
        if obj:
            return [
                ('Infos', {'fields': ['artikelnummer', 'name']}),
                ('Beschrieb', {'fields': [
                 'kurzbeschrieb', 'beschrieb'], 'classes': ["collapse start-open"]}),
                ('Daten', {'fields': [
                 'mengenbezeichnung', 'verkaufspreis', 'mwstsatz', 'lagerbestand']}),
                ('Lieferant', {'fields': [
                 'lieferant', 'lieferant_preis', 'lieferant_artikelnummer']}),
                ('Aktion', {'fields': ['aktion_von', 'aktion_bis', 'aktion_preis'], 'classes': [
                 "collapse start-open"]}),
                ('Links', {'fields': ['datenblattlink', 'bildlink'], 'classes': [
                 "collapse start-open"]}),
                # packlistenbemerkung
                ('Bemerkung / Notiz',
                 {'fields': ['bemerkung', 'html_notiz'], 'classes': ["collapse start-open"]})
            ]
        else:
            return [
                ('Infos', {'fields': ['artikelnummer', 'name']}),
                ('Beschrieb', {'fields': [
                 'kurzbeschrieb', 'beschrieb'], 'classes': ["collapse start-open"]}),
                ('Daten', {'fields': [
                 'mengenbezeichnung', 'verkaufspreis', 'mwstsatz', 'lagerbestand']}),
                ('Lieferant', {'fields': [
                 'lieferant', 'lieferant_preis', 'lieferant_artikelnummer']}),
                ('Aktion', {'fields': ['aktion_von', 'aktion_bis', 'aktion_preis'], 'classes': [
                 "collapse start-open"]}),
                ('Links', {'fields': ['datenblattlink', 'bildlink'], 'classes': [
                 "collapse start-open"]}),
                ('Bemerkung', {'fields': ['bemerkung'], 'classes': [
                 "collapse start-open"]})  # packlistenbemerkung
            ]

    ordering = ('artikelnummer', 'name')

    list_display = ('artikelnummer', 'clean_name', 'clean_kurzbeschrieb',
                    'clean_beschrieb', 'preis', 'in_aktion', 'lagerbestand', 'bild', 'html_notiz')
    list_filter = ('lieferant', 'kategorien', 'lagerbestand')
    search_fields = ['artikelnummer', 'name', 'kurzbeschrieb',
                     'beschrieb', 'bemerkung', 'notiz__name', 'notiz__beschrieb']

    readonly_fields = ["html_notiz"]

    inlines = (ProduktInlineProduktkategorien,)

    save_on_top = True

    actions = ["wc_update", "lagerbestand_zuruecksetzen", "aktion_beenden"]

    list_select_related = ["notiz"]

    def wc_update(self, request, queryset):
        result = WooCommerce.product_bulk_update(queryset.all())
        messages.success(request, ((
            '{} Produkte' if result[0] != 1 else 'Ein Produkt') + ' von WooCommerce aktualisiert!').format(result[0]))
        if result[1]:
            messages.error(request, ('Beim Import von ' + (
                '{} Produkten' if result[1] != 1 else 'einem Produkt') + ' von WooCommerce ist ein Fehler aufgetreten!').format(result[1]))
    wc_update.short_description = "Produkte von WooCommerce aktualisieren"

    def lagerbestand_zuruecksetzen(self, request, queryset):
        for produkt in queryset.all():
            produkt.lagerbestand = 0
            produkt.save()
        messages.success(request, (('Lagerbestand von {} Produkten' if queryset.count(
        ) != 1 else 'Lagerbestand von einem Produkt') + ' zurückgesetzt!').format(queryset.count()))
    lagerbestand_zuruecksetzen.short_description = "Lagerbestand zurücksetzen"

    def aktion_beenden(self, request, queryset):
        for produkt in queryset.all():
            produkt.aktion_bis = datetime.now(utc)
            produkt.save()
        messages.success(request, (('Aktion von {} Produkten' if queryset.count(
        ) != 1 else 'Aktion von einem Produkt') + ' beendet!').format(queryset.count()))
    aktion_beenden.short_description = "Aktion beenden"


@admin.register(Zahlungsempfaenger)
class ZahlungsempfaengerAdmin(admin.ModelAdmin):
    fieldsets = [
        ("Infos", {"fields": [
         "qriban", "logourl", "firmenname", "firmenuid"]}),
        ("Adresse", {"fields": ["adresszeile1", "adresszeile2", "land"]}),
        ("Daten", {"fields": ["email", "telefon", "webseite"]})
    ]

    list_display = ('firmenname', 'firmenuid', 'qriban')
    list_filter = ('land',)
    search_fields = ['firmenname', 'adresszeile1',
                     'adresszeile2', 'qriban', 'firmenuid']

    save_on_top = True


###################

@admin.register(Einstellung)
class EinstellungenAdmin(admin.ModelAdmin):
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    list_display = ('name', 'get_inhalt')
    ordering = ('name',)

    search_fields = ["name"]

    readonly_fields = ["id", "name"]

    def get_fieldsets(self, request, obj=None):
        fieldsets = [
            ('Name', {'fields': ['name']}),
            ('Inhalt', {'fields': ['char']}),
        ]
        if obj.typ == "char":
            fieldsets[1][1]["fields"][0] = "char"
        elif obj.typ == "text":
            fieldsets[1][1]["fields"][0] = "text"
        elif obj.typ == "int":
            fieldsets[1][1]["fields"][0] = "inte"
        elif obj.typ == "floa":
            fieldsets[1][1]["fields"][0] = "floa"
        elif obj.typ == "url":
            fieldsets[1][1]["fields"][0] = "url"

        return fieldsets

# TO DO


@admin.register(ToDoNotiz)
class ToDoNotizenAdmin(NotizenAdmin):
    list_editable = ["priority", "erledigt"]
    list_filter = ["priority"]

    ordering = ["-priority", "erstellt_am"]

    def has_module_permission(self, request):
        return {}

    def has_delete_permission(self, request, obj=None):
        return False

    def get_urls(self):
        from django.urls import path
        from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe
        from functools import update_wrapper

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(allow_iframe(view))(*args, **kwargs)
            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        return [
            path('', wrap(self.changelist_view),
                 name='%s_%s_changelist' % info),
            path('add/', wrap(self.add_view), name='%s_%s_add' % info),
            path('autocomplete/', wrap(self.autocomplete_view),
                 name='%s_%s_autocomplete' % info),
            path('<path:object_id>/history/', wrap(self.history_view),
                 name='%s_%s_history' % info),
            # path('<path:object_id>/delete/', wrap(self.delete_view), name='%s_%s_delete' % info),
            path('<path:object_id>/change/', wrap(self.change_view),
                 name='%s_%s_change' % info),
        ]


@admin.register(ToDoVersand)
class ToDoVersandAdmin(BestellungsAdmin):
    list_display = ('id', 'info', 'trackingnummer',
                    'versendet', 'status', 'html_todo_notiz')
    list_editable = ("trackingnummer", "versendet", "status")
    list_filter = ('status', 'bezahlt')

    ordering = ("bezahlt", "-datum")

    actions = ()

    def has_module_permission(self, request):
        return {}

    def has_delete_permission(self, request, obj=None):
        return False

    def get_urls(self):
        from django.urls import path
        from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe
        from functools import update_wrapper

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(allow_iframe(view))(*args, **kwargs)
            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        return [
            path('', wrap(self.changelist_view),
                 name='%s_%s_changelist' % info),
            path('add/', wrap(self.add_view), name='%s_%s_add' % info),
            path('autocomplete/', wrap(self.autocomplete_view),
                 name='%s_%s_autocomplete' % info),
            path('<path:object_id>/history/', wrap(self.history_view),
                 name='%s_%s_history' % info),
            # path('<path:object_id>/delete/', wrap(self.delete_view), name='%s_%s_delete' % info),
            path('<path:object_id>/change/', wrap(self.change_view),
                 name='%s_%s_change' % info),
        ]


@admin.register(ToDoZahlungseingang)
class ToDoZahlungseingangAdmin(BestellungsAdmin):
    list_display = ('id', 'info', 'bezahlt', 'status',
                    'fix_summe', 'html_todo_notiz')
    list_editable = ("bezahlt", "status")
    list_filter = ('status', 'versendet', 'zahlungsmethode')

    ordering = ("versendet", "-datum")

    actions = ()

    def has_module_permission(self, request):
        return {}

    def has_delete_permission(self, request, obj=None):
        return False

    def get_urls(self):
        from django.urls import path
        from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe
        from functools import update_wrapper

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(allow_iframe(view))(*args, **kwargs)
            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        return [
            path('', wrap(self.changelist_view),
                 name='%s_%s_changelist' % info),
            path('add/', wrap(self.add_view), name='%s_%s_add' % info),
            path('autocomplete/', wrap(self.autocomplete_view),
                 name='%s_%s_autocomplete' % info),
            path('<path:object_id>/history/', wrap(self.history_view),
                 name='%s_%s_history' % info),
            # path('<path:object_id>/delete/', wrap(self.delete_view), name='%s_%s_delete' % info),
            path('<path:object_id>/change/', wrap(self.change_view),
                 name='%s_%s_change' % info),
        ]


@admin.register(ToDoLagerbestand)
class ToDoLagerbestandAdmin(ProduktAdmin):
    list_display = ('nr', 'clean_name', 'lagerbestand',
                    'preis', 'bemerkung', 'html_todo_notiz')
    list_editable = ["lagerbestand"]

    actions = ["lagerbestand_zuruecksetzen"]

    def has_module_permission(self, request):
        return {}

    def has_delete_permission(self, request, obj=None):
        return False

    def get_urls(self):
        from django.urls import path
        from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe
        from functools import update_wrapper

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(allow_iframe(view))(*args, **kwargs)
            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        return [
            path('', wrap(self.changelist_view),
                 name='%s_%s_changelist' % info),
            path('add/', wrap(self.add_view), name='%s_%s_add' % info),
            path('autocomplete/', wrap(self.autocomplete_view),
                 name='%s_%s_autocomplete' % info),
            path('<path:object_id>/history/', wrap(self.history_view),
                 name='%s_%s_history' % info),
            # path('<path:object_id>/delete/', wrap(self.delete_view), name='%s_%s_delete' % info),
            path('<path:object_id>/change/', wrap(self.change_view),
                 name='%s_%s_change' % info),
        ]


@admin.register(ToDoLieferung)
class ToDoLieferungenAdmin(LieferungenAdmin):
    list_display = ('name', 'anzahlprodukte', 'eingelagert',
                    'notiz', 'html_todo_notiz_erstellen')
    list_editable = ("eingelagert",)
    list_filter = ()

    def has_module_permission(self, request):
        return {}

    def has_delete_permission(self, request, obj=None):
        return False

    def get_urls(self):
        from django.urls import path
        from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe
        from functools import update_wrapper

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(allow_iframe(view))(*args, **kwargs)
            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name

        return [
            path('', wrap(self.changelist_view),
                 name='%s_%s_changelist' % info),
            path('add/', wrap(self.add_view), name='%s_%s_add' % info),
            path('autocomplete/', wrap(self.autocomplete_view),
                 name='%s_%s_autocomplete' % info),
            path('<path:object_id>/history/', wrap(self.history_view),
                 name='%s_%s_history' % info),
            # path('<path:object_id>/delete/', wrap(self.delete_view), name='%s_%s_delete' % info),
            path('<path:object_id>/change/', wrap(self.change_view),
                 name='%s_%s_change' % info),
        ]



from io import BytesIO

from reportlab.graphics import renderPDF
from reportlab.graphics.barcode import qr
from reportlab.graphics.shapes import Drawing
from reportlab.lib.colors import green, white, black
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import mm, cm
from reportlab.lib.utils import ImageReader
from reportlab.pdfbase.pdfmetrics import registerFont
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfgen import canvas
from reportlab.platypus import Table, TableStyle, Paragraph, Spacer, BaseDocTemplate, SimpleDocTemplate, Frame, PageTemplate, TopPadder, Flowable

from ..utils import clean, formatprice, runden

from django.utils import translation
_ = translation.gettext

#####

class PriceTable(Table):
    def __init__(self, bestellung):
        sprache = bestellung.kunde.sprache if bestellung.kunde and bestellung.kunde.sprache else "de"

        data = [(_("Art-Nr."),_("Bezeichnung"),_("Anzahl"),_("Einheit"),_("Preis"),_("Total"))]

        style_default = ParagraphStyle("Normal",fontname="Helvetica")
        style_bold = ParagraphStyle("Bold",fontname="Helvetica-Bold")


        for bp in bestellung.produkte.through.objects.filter(bestellung=bestellung):  # Produkte
            data.append((
                bp.produkt.artikelnummer,
                Paragraph(clean(bp.produkt.name, sprache), style_default),
                str(bp.menge),
                clean(bp.produkt.mengenbezeichnung, sprache),
                formatprice(bp.produktpreis),
                formatprice(bp.zwischensumme())
            ))
            if bp.bemerkung:
                data.append((
                    "",
                    Paragraph(bp.bemerkung, style_bold),
                    "",
                    "",
                    "",
                    ""
                ))

        kostenzeilen = 0

        for bk in bestellung.kosten.through.objects.filter(bestellung=bestellung):  # Kosten
            data.append((
                "",
                Paragraph(clean(bk.kosten.name, sprache), style_default),
                "",
                "",
                "",
                formatprice(bk.kosten.preis)
            ))
            kostenzeilen += 1
            if bk.bemerkung:
                data.append((
                    "",
                    Paragraph(bk.bemerkung, style_bold),
                    "",
                    "",
                    "",
                    ""
                ))
                kostenzeilen += 1

        mwstdict = dict(bestellung.mwstdict())
        for mwstsatz in mwstdict:  # Mehrwertsteuer
            data.append((
                "",
                _("MwSt"),
                mwstsatz,
                "%",
                formatprice(float(mwstdict[mwstsatz])),
                formatprice(runden(float(mwstdict[mwstsatz]*(float(mwstsatz)/100))))
            ))
            kostenzeilen += 1

        data.append((  # Total
            _("RECHNUNGSBETRAG"),
            "",
            "",
            "CHF",
            "",
            formatprice(bestellung.summe_gesamt())
        ))

        style = [
            ('LINEABOVE', (0,0), (-1,0), 1, black),
            ('LINEBELOW', (0,0), (-1,0), 1, black),
            ('LINEABOVE', (0,-1-kostenzeilen), (-1,-1-kostenzeilen), 0.5, black),
            ('LINEABOVE', (0,-1), (-1,-1), 1, black),
            ('LINEBELOW', (0,-1), (-1,-1), 1, black),
            ('ALIGN', (-1, 0), (-1, -1), "RIGHT"),
            ('ALIGN', (-2, 0), (-2, -1), "RIGHT"),
            ('ALIGN', (1, -1), (1, -1), "CENTER"),
            ('FONTNAME', (0, -1), (-1, -1), "Helvetica-Bold"),
            ('VALIGN', (0, 0), (-1, -1), "TOP"),
        ]

        Table.__init__(self, data, repeatRows=1, style=TableStyle(style), colWidths=[26*mm,80*mm,20*mm,20*mm,20*mm,20*mm])

class ProductTable(Table):
    def __init__(self, bestellung):
        sprache = bestellung.kunde.sprache if bestellung.kunde and bestellung.kunde.sprache else "de"

        data = [(_("Art-Nr."),_("Bezeichnung"),_("Anzahl"),_("Einheit"))]

        style_default = ParagraphStyle("Normal",fontname="Helvetica")
        style_bold = ParagraphStyle("Bold",fontname="Helvetica-Bold")

        produktanzahl = 0

        for bp in bestellung.produkte.through.objects.filter(bestellung=bestellung):  # Produkte
            data.append((
                bp.produkt.artikelnummer,
                Paragraph(clean(bp.produkt.name, sprache), style_default),
                str(bp.menge),
                clean(bp.produkt.mengenbezeichnung, sprache),
            ))
            if bp.bemerkung:
                data.append((
                    "",
                    Paragraph(bp.bemerkung, style_bold),
                    "",
                    ""
                ))

            produktanzahl += bp.menge

        data.append((  # Total
            _("ANZAHL PRODUKTE"),
            "",
            str(produktanzahl),
            ""
        ))

        style = [
            ('LINEABOVE', (0,0), (-1,0), 1, black),
            ('LINEBELOW', (0,0), (-1,0), 1, black),
            ('LINEABOVE', (0,-1), (-1,-1), 1, black),
            ('LINEBELOW', (0,-1), (-1,-1), 1, black),
            ('ALIGN', (1, -1), (1, -1), "CENTER"),
            ('FONTNAME', (0, -1), (-1, -1), "Helvetica-Bold"),
            ('VALIGN', (0, 0), (-1, -1), "TOP"),
        ]

        Table.__init__(self, data, repeatRows=1, style=TableStyle(style), colWidths=[36*mm,110*mm,20*mm,20*mm])

class QrInvoice(Flowable):
    def __init__(self, bestellung, digital=True):
        Flowable.__init__(self)
        self.width = 210
        self.height = 110
        self._fixedWidth = 210
        self._fixedHeight = 110
        self.digital = digital
        self.bestellung = bestellung

    def __repr__(self):
        return "QR-Invoice"

    def __str__(self):
        return "QR-Invoice"

    def debug(self):
        c = self.canv
        c.setStrokeColor("green")
        c.rect(5*mm,5*mm,52*mm,95*mm) # Empfangsschein
        c.rect(67*mm,5*mm,138*mm,95*mm) # Zahlteil
        c.rect(67*mm,42*mm,46*mm,46*mm) # QR-Code
        c.line(5*mm,23*mm,57*mm,23*mm)
        c.line(5*mm,37*mm,57*mm,37*mm)
        c.line(5*mm,93*mm,57*mm,93*mm)
        c.line(67*mm,93*mm,118*mm,93*mm)
        c.line(67*mm,37*mm,118*mm,37*mm)
        c.line(67*mm,15*mm,205*mm,15*mm)
        c.line(118*mm,100*mm,118*mm,15*mm)
        c.setStrokeColor("black")

    def get_swiss_qr_payload(self):
        bestellung = self.bestellung
        ze = bestellung.zahlungsempfaenger
        qrpayload = []

        def ln(text=""):
            qrpayload.append(text)

        #QRCH
            #Header
                #QRType
        ln("SPC")
                #Version
        ln("0200")
                #Coding
        ln("1")
            #CdtrInf (Empfänger)
                #IBAN
        ln(ze.qriban.replace(" ",""))
                #Cdtr
                    #AdrTp
        ln("K")
                    #Name
        ln(ze.firmenname)
                    #StrtNmOrAdrLine1
        ln(ze.adresszeile1)
                    #BldgNbOrAdrLine2
        ln(ze.adresszeile2)
                    #PstCd
        ln()
                    #TwnNm
        ln()
                    #Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(ze.land)
            #UltmtCdtr (Entgültiger Zahlungsempfänger)
                #AdrTp
        ln()
                #Name
        ln()
                #StrtNmOrAdrLine1
        ln()
                #BldgNbOrAdrLine2
        ln()
                #PstCd
        ln()
                #TwnNm
        ln()
                #Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln()
            #CcyAmt
                #Amt
        ln("{:.2f}".format(bestellung.summe_gesamt()))
                #Ccy
        ln("CHF")
            #UltmtDbtr (Entgültiger Zahlungspflichtiger)
                #AdrTp
        ln("K")
                #Name
        ln((bestellung.rechnungsadresse_vorname+" "+bestellung.rechnungsadresse_nachname) if not bestellung.rechnungsadresse_firma else bestellung.rechnungsadresse_firma)
                #StrtNmOrAdrLine1
        ln(bestellung.rechnungsadresse_adresszeile1)
                #BldgNbOrAdrLine2
        ln(bestellung.rechnungsadresse_plz+" "+bestellung.rechnungsadresse_ort)
                #PstCd
        ln()
                #TwnNm
        ln()
                #Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(bestellung.rechnungsadresse_land)
            #RmtIn
                #TP
        ln("QRR")
                #Ref
        ln(bestellung.referenznummer().replace(" ",""))
                #AddInf
                    #Ustrd
        ln(str(bestellung.datum.strftime("%d.%m.%Y")))
                    #Trailer
        ln("EPD")
                    #StrdBkgInf
        ln(bestellung.rechnungsinformationen())
            #AltPmtInf
                #AltPmt
        #ln()
                #AltPmt
        #ln()

        return "\n".join(qrpayload)

    def draw_qr_invoice(self):
        bestellung = self.bestellung
        bestelldatum = str(bestellung.datum.strftime("%d.%m.%Y"))
        rechnungsinformationen = bestellung.rechnungsinformationen().split("/31/")
        referenznummer = bestellung.referenznummer()
        gesamtsumme = format(bestellung.summe_gesamt(),"08,.2f").replace(","," ").lstrip(" 0")
        ze = bestellung.zahlungsempfaenger

        c = self.canv
        c.saveState()

        # QR-Code

        qrpayload = self.get_swiss_qr_payload()
        qr_code = qr.QrCodeWidget(qrpayload)
        qrbounds = qr_code.getBounds()
        qrwidth = qrbounds[2] - qrbounds[0]
        qrheight = qrbounds[3] - qrbounds[1]
        d = Drawing(52.2*mm, 52.2*mm, transform=[52.2*mm/qrwidth,0,0,52.2*mm/qrheight,0,0]) # 46, 46
        d.add(qr_code)
        renderPDF.draw(d, c, 63.9*mm, 38.9*mm) # 67, 42

        # Schweizerkreuz

        c.setFillColor("black")
        c.setStrokeColor("white")
        c.rect(86.5*mm,61.5*mm,7*mm,7*mm,fill=1,stroke=1)
        c.setFillColor("white")
        c.rect(89.25*mm,63*mm,1.5*mm,4*mm,fill=1,stroke=0)
        c.rect(88*mm,64.25*mm,4*mm,1.5*mm,fill=1,stroke=0)

        c.setFillColor("black")
        c.setStrokeColor("black")

        # Begrenzungen Empfangsschein und Zahlteil und Abzutrennen-Hinweis

        if self.digital:
            c.line(0*mm,105*mm,210*mm,105*mm)
            c.line(62*mm,0*mm,62*mm,105*mm)
            c.setFont("Helvetica-Bold", 8)
            c.drawCentredString(105*mm, 107*mm, _("Vor der Einzahlung abzutrennen"))

        # Titel

        def titel(t, text, klein=False):
            t.setFont("Helvetica-Bold", 6 if klein else 8)
            t.textLine(text)
            t.moveCursor(0,2)
            t.setFont("Helvetica", 8 if klein else 10)

        # Empfangsschein Angaben
        t = c.beginText(5*mm,90*mm)
        titel(t,_("Konto / Zahlbar an"),True)
        t.textLine(ze.qriban)
        t.textLine(ze.firmenname)
        t.textLine(ze.adresszeile1)
        t.textLine(ze.adresszeile2)
        t.moveCursor(0,9)
        titel(t,_("Referenz"),True)
        t.textLine(referenznummer)
        t.moveCursor(0,9)
        titel(t,_("Zahlbar durch"),True)
        t.textLine((bestellung.rechnungsadresse_vorname+" "+bestellung.rechnungsadresse_nachname) if not bestellung.rechnungsadresse_firma else bestellung.rechnungsadresse_firma)
        t.textLine(bestellung.rechnungsadresse_adresszeile1)
        t.textLine(bestellung.rechnungsadresse_plz+" "+bestellung.rechnungsadresse_ort)
        c.drawText(t)

        # Zahlteil Angaben
        t = c.beginText(118*mm,97*mm)
        titel(t,_("Konto / Zahlbar an"))
        t.textLine(ze.qriban)
        t.textLine(ze.firmenname)
        t.textLine(ze.adresszeile1)
        t.textLine(ze.adresszeile2)
        t.moveCursor(0,9)
        titel(t,_("Referenz"))
        t.textLine(referenznummer)
        t.moveCursor(0,9)
        titel(t,_("Zusätzliche Informationen"))
        t.textLine(bestelldatum)
        t.textLine(rechnungsinformationen[0])
        t.textLine("/31/"+rechnungsinformationen[1])
        t.moveCursor(0,9)
        titel(t,_("Zahlbar durch"))
        t.textLine((bestellung.rechnungsadresse_vorname+" "+bestellung.rechnungsadresse_nachname) if not bestellung.rechnungsadresse_firma else bestellung.rechnungsadresse_firma)
        t.textLine(bestellung.rechnungsadresse_adresszeile1)
        t.textLine(bestellung.rechnungsadresse_plz+" "+bestellung.rechnungsadresse_ort)
        c.drawText(t)

        # Texte
        c.setFont("Helvetica-Bold", 11)
        c.drawString(5*mm, 97*mm, _("Empfangsschein"))
        c.drawString(67*mm, 97*mm, _("Zahlteil"))

        c.setFont("Helvetica-Bold", 6)
        c.drawString(5*mm, 33*mm, _("Währung"))
        c.drawString(20*mm, 33*mm, _("Betrag"))
        c.drawString(38*mm, 20*mm, _("Annahmestelle"))

        c.setFont("Helvetica", 8)
        c.drawString(5*mm, 30*mm, "CHF")
        c.drawString(20*mm, 30*mm, gesamtsumme)

        c.setFont("Helvetica-Bold", 8)
        c.drawString(67*mm, 33*mm, _("Währung"))
        c.drawString(87*mm, 33*mm, _("Betrag"))

        c.setFont("Helvetica", 10)
        c.drawString(67*mm, 29*mm, "CHF")
        c.drawString(87*mm, 29*mm, gesamtsumme)

        # c.setFont("Helvetica-Bold", 7)
        # c.drawString(67*mm, 11*mm, "Name AV1:")
        # c.drawString(67*mm, 8*mm, "Name AV2:")
        #
        # c.setFont("Helvetica", 7)
        # c.drawString(82*mm, 11*mm, "Linie 1")
        # c.drawString(82*mm, 8*mm, "Linie 2")

        # if settings.DEBUG:
            # self.debug()

        c.restoreState()

    def draw(self):
        self.canv.translate(-12*mm,-12*mm)
        self.draw_qr_invoice()

class Header(Flowable):
    def __init__(self, bestellung, lieferschein=False):
        Flowable.__init__(self)
        self.width = 210
        self.height = 75
        self._fixedWidth = 210
        self._fixedHeight = 75
        self.lieferschein = lieferschein
        self.bestellung = bestellung

    def __repr__(self):
        return "QR-Invoice"

    def __str__(self):
        return "QR-Invoice"

    def draw_header(self):
        bestellung = self.bestellung
        bestelldatum = str(bestellung.datum.strftime("%d.%m.%Y"))
        ze = bestellung.zahlungsempfaenger

        c = self.canv
        c.saveState()

        # Logo
        if ze.logourl:
            c.drawImage(ImageReader(ze.logourl), 120*mm, 67*mm, width=20*mm, height=-20*mm, mask="auto", anchor="nw")

        # Firmenname
        c.setFont("Helvetica-Bold", 14)
        c.drawString(12*mm, 64*mm, ze.firmenname)

        # Firmenadresse
        t = c.beginText(12*mm, 57*mm)
        t.setFont("Helvetica", 10)
        t.textLine(ze.adresszeile1)
        t.textLine((ze.land+"-" if not ze.land == "CH" else "")+ze.adresszeile2)
        c.drawText(t)

        # Firmendaten: Texte
        t = c.beginText(12*mm, 46*mm)
        t.setFont("Helvetica", 8)
        t.textLine(_("Tel."))
        t.textLine(_("E-Mail"))
        t.textLine(_("Web"))
        t.textLine(_("MwSt"))
        c.drawText(t)

        # Firmendaten: Inhalt
        t = c.beginText(24*mm, 46*mm)
        t.setFont("Helvetica", 8)
        t.textLine(bestellung.ansprechpartner.telefon)
        t.textLine(bestellung.ansprechpartner.email)
        t.textLine(ze.webseite)
        t.textLine(ze.firmenuid)
        c.drawText(t)

        # Rechnungsdaten: Texte
        t = c.beginText(12*mm, 27*mm)
        t.setFont("Helvetica", 12)
        t.textLine(_("Ihr/e Ansprechpartner/in"))
        t.textLine(_("Ihre Bestellung vom"))
        t.textLine(_("Ihre Kundennummer"))
        c.drawText(t)

        # Rechnungsdaten: Inhalt
        t = c.beginText(64*mm, 27*mm)
        t.setFont("Helvetica", 12)
        t.textLine(bestellung.ansprechpartner.name)
        t.textLine(bestelldatum)
        t.textLine(str(bestellung.kunde.pk).zfill(6) if bestellung.kunde else "n.a.")
        c.drawText(t)

        # Kundenadresse
        t = c.beginText(120*mm, 27*mm)
        t.setFont("Helvetica", 12)
        if self.lieferschein:
            if bestellung.lieferadresse_firma:
                t.textLine(bestellung.lieferadresse_firma)
            if bestellung.lieferadresse_vorname or bestellung.lieferadresse_nachname:
                t.textLine((bestellung.lieferadresse_vorname+" " if bestellung.lieferadresse_vorname else "")+(bestellung.lieferadresse_nachname or ""))
            t.textLine(bestellung.lieferadresse_adresszeile1)
            if bestellung.lieferadresse_adresszeile2:
                t.textLine(bestellung.lieferadresse_adresszeile2)
            t.textLine(bestellung.lieferadresse_plz+" "+bestellung.lieferadresse_ort)
            c.drawText(t)
        else:
            if bestellung.rechnungsadresse_firma:
                t.textLine(bestellung.rechnungsadresse_firma)
            if bestellung.rechnungsadresse_vorname or bestellung.rechnungsadresse_nachname:
                t.textLine((bestellung.rechnungsadresse_vorname+" " if bestellung.rechnungsadresse_vorname else "")+(bestellung.rechnungsadresse_nachname or ""))
            t.textLine(bestellung.rechnungsadresse_adresszeile1)
            if bestellung.rechnungsadresse_adresszeile2:
                t.textLine(bestellung.rechnungsadresse_adresszeile2)
            t.textLine(bestellung.rechnungsadresse_plz+" "+bestellung.rechnungsadresse_ort)
            c.drawText(t)

        # Rechnung
        c.setFont("Helvetica-Bold", 10)
        if self.lieferschein:
            c.drawString(12*mm, 0*mm, _("LIEFERSCHEIN"))
        else:
            c.drawString(12*mm, 0*mm, _("RECHNUNG"))
        c.setFont("Helvetica", 10)
        c.drawString(64*mm, 0*mm, (bestellung.datum.strftime("%Y")+"-"+str(bestellung.pk).zfill(6)+(" (Online #"+str(bestellung.woocommerceid)+")" if bestellung.woocommerceid else "")))

        c.restoreState()

    def draw(self):
        self.canv.translate(-12*mm,-40*mm)
        self.draw_header()

def pdf_bestellung(bestellung, lieferschein=False, digital:bool=True):
    sprache = bestellung.kunde.sprache if bestellung.kunde and bestellung.kunde.sprache else "de"

    cur_language = translation.get_language()
    translation.activate(sprache)

    #####

    def get_buffer(bestellung, lieferschein, digital):
        buffer = BytesIO()

        doc = SimpleDocTemplate(buffer, rightMargin=10*mm, leftMargin=10*mm, topMargin=10*mm, bottomMargin=10*mm)
        if lieferschein:
            elements = [Header(bestellung, lieferschein=True), Spacer(1,48*mm), ProductTable(bestellung)]
        else:
            elements = [Header(bestellung, lieferschein=False), Spacer(1,48*mm), PriceTable(bestellung), Spacer(1,65*mm), TopPadder(QrInvoice(bestellung, digital))]
        doc.build(elements)
        buffer.seek(0)
        return buffer

    # def get_buffer_old(bestellung):
    #     buffer = BytesIO()
    #
    #     template = PageTemplate(id="empty", pagesize=A4, frames=[Frame(10*mm, 10*mm, 190*mm, 277*mm)])
    #     doc = BaseDocTemplate(buffer, pagesize=A4, pageTemplates=[template])
    #     elements = [Header(bestellung, lieferschein=True), Spacer(1,48*mm), PriceTable(bestellung), Spacer(1,65*mm), TopPadder(QrInvoice(bestellung, digital))]
    #     doc.build(elements)
    #     buffer.seek(0)
    #     return buffer

    #####

    buffer = get_buffer(bestellung, lieferschein=lieferschein, digital=digital)

    translation.activate(cur_language)

    return buffer



from django.db import models
from django.core import mail
from django.conf import settings
from django.core.validators import RegexValidator
from django.http import FileResponse
from django.template.loader import get_template
from django.utils.html import mark_safe
from django.urls import reverse

from datetime import datetime
from random import randint
from pytz import utc
from django.utils import timezone

from .utils import send_mail, runden, clean, formatprice, modulo10rekursiv, send_pdf
from .pdf_generators.bestellung import pdf_bestellung

###################

def defaultlieferungsname():
    return "Lieferung vom "+str(datetime.now().strftime("%d.%m.%Y"))

def defaultbestellungsname():
    return "Bestellung vom "+str(datetime.now().strftime("%d.%m.%Y"))

def defaultzahlungsempfaenger():
    return Zahlungsempfaenger.objects.first().pk

def defaultansprechpartner():
    return Ansprechpartner.objects.first().pk

def defaultorderkey():
    return "kh-"+str(randint(10000000,99999999))


STATUS = [
    ("pending","Zahlung ausstehend"),
    ("processing","In Bearbeitung"),
    ("on-hold","In Wartestellung"),
    ("completed","Abgeschlossen"),
    ("cancelled","Storniert/Abgebrochen"),
    ("refunded","Rückerstattet"),
    ("failed","Fehlgeschlagen"),
    ("trash","Gelöscht")
]

MWSTSÄTZE = [
(0.0, "0.0% (Mehrwertsteuerfrei)"),
(7.7, "7.7% (Normalsatz)"),
(3.7, "3.7% (Sondersatz für Beherbergungsdienstleistungen)"),
(2.5, "2.5% (Reduzierter Satz)")
]

ZAHLUNGSMETHODEN = [
    ("bacs",    "Überweisung"),
    ("cheque",  "Scheck"),
    ("cod",     "Rechnung / Nachnahme"),
    ("paypal",  "PayPal")
]

LÄNDER = [
    ("CH","Schweiz"),
    ("LI","Liechtenstein")
]

SPRACHEN = [
    ("de","Deutsch [DE]"),
    ("fr","Französisch [FR]"),
    ("it","Italienisch [IT]"),
    ("en","Englisch [EN]")
]

GUTSCHEINTYPEN = [
    ("percent",         "Prozent"),
    ("fixed_cart",      "Fixer Betrag auf den Warenkorb"),
    ("fixed_product",   "Fixer Betrag auf ein Produkt")
]

#############

class Ansprechpartner(models.Model):
    name = models.CharField('Name', max_length=50, help_text="Auf Rechnung ersichtlich!")
    telefon = models.CharField('Telefon', max_length=50, help_text="Auf Rechnung ersichtlich!")
    email = models.EmailField('E-Mail', help_text="Auf Rechnung ersichtlich!")

    def __str__(self):
        return self.name

    __str__.short_description = 'Ansprechpartner'

    class Meta:
        verbose_name = "Ansprechpartner"
        verbose_name_plural = "Ansprechpartner"



class Bestellungskosten(models.Model):
    bestellung = models.ForeignKey("Bestellung", on_delete=models.CASCADE)
    kosten = models.ForeignKey("Kosten", on_delete=models.PROTECT)
    bemerkung = models.CharField("Bemerkung", default="", max_length=250, blank=True, help_text="Wird auf die Rechnung gedruckt.")

    def zwischensumme(self):
        return runden(self.kosten.preis)
    zwischensumme.short_description = "Zwischensumme (exkl. MwSt) in CHF"

    def mwstsatz(self):
        return formatprice(self.kosten.mwstsatz)
    mwstsatz.short_description = "MwSt-Satz"

    def __str__(self):
        return "1x "+str(self.kosten)
    __str__.short_description = "Bestellungskosten"

    class Meta:
        verbose_name = "Bestellungskosten"
        verbose_name_plural = "Bestellungskosten"

class Bestellungsposten(models.Model):
    bestellung = models.ForeignKey("Bestellung", on_delete=models.CASCADE)
    produkt = models.ForeignKey("Produkt", on_delete=models.PROTECT)
    bemerkung = models.CharField("Bemerkung", default="", max_length=250, blank=True, help_text="Wird auf die Rechnung gedruckt.")
    menge = models.IntegerField("Menge", default=1)

    produktpreis = models.FloatField("Produktpreis in CHF (exkl. MwSt)", default=0.0)

    def zwischensumme(self):
        return runden(self.produktpreis*self.menge)
    zwischensumme.short_description = "Zwischensumme (exkl. MwSt) in CHF"

    def mwstsatz(self):
        return formatprice(self.produkt.mwstsatz)
    mwstsatz.short_description = "MwSt-Satz"

    def __str__(self):
        return str(self.menge)+"x "+self.produkt.clean_name()
    __str__.short_description = "Bestellungsposten"

    def save(self, *args, **kwargs):
        if not self.produktpreis:
            self.produktpreis = runden(self.produkt.preis())
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Bestellungsposten"
        verbose_name_plural = "Bestellungsposten"

class Bestellung(models.Model):
    woocommerceid = models.IntegerField('WooCommerce ID', default=0)

    datum = models.DateTimeField("Datum", default=timezone.now)

    status = models.CharField("Status", max_length=11, default="pending", choices=STATUS)
    versendet = models.BooleanField("Versendet", default=False, help_text="Sobald eine Bestellung als versendet markiert wurde, kann sie nicht mehr bearbeitet werden! (Ausgenommen Status/Bezahlt) Ausserdem werden die Produkte aus dem Lagerbestand entfernt.")
    trackingnummer = models.CharField("Trackingnummer", default="", blank=True, max_length=25, validators=[RegexValidator(r'^99\.[0-9]{2}\.[0-9]{6}\.[0-9]{8}$', 'Bite benutze folgendes Format: 99.xx.xxxxxx.xxxxxxxx')], help_text="Bitte gib hier eine Trackingnummer der Schweizer Post ein. (optional)")

    ausgelagert = models.BooleanField("Ausgelagert", default=False)

    zahlungsmethode = models.CharField("Zahlungsmethode", max_length=7, default="cod", choices=ZAHLUNGSMETHODEN)
    bezahlt = models.BooleanField("Bezahlt", default=False, help_text="Sobald eine Bestellung als bezahlt markiert wurde, kann sie nicht mehr bearbeitet werden! (Ausgenommen Status/Versendet/Trackingnummer)")

    kundennotiz = models.TextField("Kundennotiz", default="", blank=True, help_text="Vom Kunden erfasste Notiz.")
    #rechnungsnotiz = models.TextField("Rechnungsnotiz", default="", blank=True, help_text="Wird auf der Rechnung gedruckt.")

    order_key = models.CharField("Bestellungs-Schlüssel", max_length=50, default=defaultorderkey, blank=True)

    kunde = models.ForeignKey("Kunde", on_delete=models.SET_NULL, null=True, blank=True)
    zahlungsempfaenger = models.ForeignKey("Zahlungsempfaenger", on_delete=models.PROTECT, verbose_name="Zahlungsempfänger", default=defaultzahlungsempfaenger)
    ansprechpartner = models.ForeignKey("Ansprechpartner", on_delete=models.PROTECT, verbose_name="Ansprechpartner", default=defaultansprechpartner)

    rechnungsadresse_vorname = models.CharField("Vorname", max_length=50, default="", blank=True)
    rechnungsadresse_nachname = models.CharField("Nachname", max_length=50, default="", blank=True)
    rechnungsadresse_firma = models.CharField("Firma", max_length=50, default="", blank=True)
    rechnungsadresse_adresszeile1 = models.CharField("Adresszeile 1", max_length=50, default="", blank=True, help_text='Strasse und Hausnummer oder "Postfach" ohne Nummer - Wird bei QR-Rechnung als Strasse und Hausnummer bzw. Postfach verwendet!')
    rechnungsadresse_adresszeile2 = models.CharField("Adresszeile 2", max_length=50, default="", blank=True, help_text="Wird in QR-Rechnung NICHT verwendet!")
    rechnungsadresse_ort = models.CharField("Ort", max_length=50, default="", blank=True)
    rechnungsadresse_kanton = models.CharField("Kanton", max_length=50, default="", blank=True)
    rechnungsadresse_plz = models.CharField("Postleitzahl", max_length=50, default="", blank=True)
    rechnungsadresse_land = models.CharField("Land", max_length=2, default="CH", choices=LÄNDER)
    rechnungsadresse_email = models.EmailField("E-Mail Adresse", blank=True)
    rechnungsadresse_telefon = models.CharField("Telefon", max_length=50, default="", blank=True)

    lieferadresse_vorname = models.CharField("Vorname", max_length=50, default="", blank=True)
    lieferadresse_nachname = models.CharField("Nachname", max_length=50, default="", blank=True)
    lieferadresse_firma = models.CharField("Firma", max_length=50, default="", blank=True)
    lieferadresse_adresszeile1 = models.CharField("Adresszeile 1", max_length=50, default="", blank=True)
    lieferadresse_adresszeile2 = models.CharField("Adresszeile 2", max_length=50, default="", blank=True)
    lieferadresse_ort = models.CharField("Ort", max_length=50, default="", blank=True)
    lieferadresse_kanton = models.CharField("Kanton", max_length=50, default="", blank=True)
    lieferadresse_plz = models.CharField("Postleitzahl", max_length=50, default="", blank=True)
    lieferadresse_land = models.CharField("Land", max_length=2, default="CH", choices=LÄNDER)

    produkte = models.ManyToManyField("Produkt", through="Bestellungsposten", through_fields=("bestellung","produkt"))

    kosten = models.ManyToManyField("Kosten", through="Bestellungskosten", through_fields=("bestellung","kosten"))

    rechnung_gesendet = models.BooleanField("Rechnung gesendet", default=False)

    fix_summe = models.FloatField("Summe in CHF", default=0.0)

    def save(self, *args, **kwargs):
        double_save = True
        if self.pk:
            self.fix_summe = self.summe_gesamt()
            double_save = False
        if self.pk is None and not self.woocommerceid and self.kunde:
            self.rechnungsadresse_vorname = self.kunde.rechnungsadresse_vorname
            self.rechnungsadresse_nachname = self.kunde.rechnungsadresse_nachname
            self.rechnungsadresse_firma = self.kunde.rechnungsadresse_firma
            self.rechnungsadresse_adresszeile1 = self.kunde.rechnungsadresse_adresszeile1
            self.rechnungsadresse_adresszeile2 = self.kunde.rechnungsadresse_adresszeile2
            self.rechnungsadresse_ort = self.kunde.rechnungsadresse_ort
            self.rechnungsadresse_kanton = self.kunde.rechnungsadresse_kanton
            self.rechnungsadresse_plz = self.kunde.rechnungsadresse_plz
            self.rechnungsadresse_land = self.kunde.rechnungsadresse_land
            self.rechnungsadresse_email = self.kunde.rechnungsadresse_email
            self.rechnungsadresse_telefon = self.kunde.rechnungsadresse_telefon

            self.lieferadresse_vorname = self.kunde.lieferadresse_vorname
            self.lieferadresse_nachname = self.kunde.lieferadresse_nachname
            self.lieferadresse_firma = self.kunde.lieferadresse_firma
            self.lieferadresse_adresszeile1 = self.kunde.lieferadresse_adresszeile1
            self.lieferadresse_adresszeile2 = self.kunde.lieferadresse_adresszeile2
            self.lieferadresse_ort = self.kunde.lieferadresse_ort
            self.lieferadresse_kanton = self.kunde.lieferadresse_kanton
            self.lieferadresse_plz = self.kunde.lieferadresse_plz
            self.lieferadresse_land = self.kunde.lieferadresse_land
        if self.versendet and (not self.ausgelagert):
            for i in self.produkte.through.objects.filter(bestellung=self):
                i.produkt.lagerbestand -= i.menge
                i.produkt.save()
            self.ausgelagert = True
        super().save(*args, **kwargs)
        if double_save:
            self.save()

    def trackinglink(self):
        return "https://www.post.ch/swisspost-tracking?formattedParcelCodes="+self.trackingnummer if self.trackingnummer else None
    trackinglink.short_description = "Trackinglink"

    def referenznummer(self):
        a = str(self.pk).zfill(22)+"0000"
        b = a+str(modulo10rekursiv(a))
        c = b[0:2]+" "+b[2:7]+" "+b[7:12]+" "+b[12:17]+" "+b[17:22]+" "+b[22:27]
        return c

    def rechnungsinformationen(self):
        date = self.datum.strftime("%y%m%d")
        uid = self.zahlungsempfaenger.firmenuid.split("-")[1].replace(".","")
        mwstdict = self.mwstdict()
        mwststring = ";".join(satz+":"+str(mwstdict[satz]) for satz in mwstdict)
        return "//S1/10/"+str(self.pk)+"/11/"+date+"/30/"+uid+"/31/"+date+"/32/"+mwststring+"/40/"+"0:30"

    def mwstdict(self):
        mwst = {}
        for p in self.produkte.through.objects.filter(bestellung=self):
            if str(p.produkt.mwstsatz) in mwst:
                mwst[str(p.produkt.mwstsatz)] += p.zwischensumme()
            else:
                mwst[str(p.produkt.mwstsatz)] = p.zwischensumme()
        for k in self.kosten.through.objects.filter(bestellung=self):
            if str(k.kosten.mwstsatz) in mwst:
                mwst[str(k.kosten.mwstsatz)] += k.zwischensumme()
            else:
                mwst[str(k.kosten.mwstsatz)] = k.zwischensumme()
        return mwst

    def summe(self):
        summe = 0
        for i in self.produkte.through.objects.filter(bestellung=self):
            summe += i.zwischensumme()
        for i in self.kosten.through.objects.filter(bestellung=self):
            summe += i.zwischensumme()
        return runden(summe)
    summe.short_description = "Summe (exkl. MwSt) in CHF"

    def summe_mwst(self):
        summe_mwst = 0
        mwstdict = self.mwstdict()
        for mwstsatz in mwstdict:
            summe_mwst += runden(float(mwstdict[mwstsatz]*(float(mwstsatz)/100)))
        return runden(summe_mwst)
    summe_mwst.short_description = "Summe (nur MwSt) in CHF"

    def summe_gesamt(self):
        return runden(self.summe()+self.summe_mwst())
    summe_gesamt.short_description = "Summe in CHF"

    def name(self):
        return (self.datum if self.datum else datetime.today()).strftime("%Y")+"-"+str(self.pk).zfill(6)+(" (WC#"+str(self.woocommerceid)+")" if self.woocommerceid else "")+" - "+(str(self.kunde) if self.kunde is not None else "Gast")
    name.short_description = "Name"

    def info(self):
        return self.datum.strftime("%d.%m.%Y")+" - "+((self.kunde.firma if self.kunde.firma else (self.kunde.vorname+" "+self.kunde.nachname)) if self.kunde else "Gast")
    info.short_description = "Info"

    def __str__(self):
        return self.name()
    __str__.short_description = "Bestellung"

    def get_pdf(self, lieferschein:bool=False, digital:bool=True):
        return FileResponse(pdf_bestellung(self, lieferschein=lieferschein, digital=digital), as_attachment=False, filename=('Lieferschein' if lieferschein else 'Rechnung')+' zu Bestellung '+str(self)+'.pdf')

    def send_pdf_rechnung_to_customer(self):
        success = send_pdf(
            subject="Ihre Rechnung Nr. "+str(self.id)+(" (Online #"+str(self.woocommerceid)+")" if self.woocommerceid else ""),
            to=self.rechnungsadresse_email,
            template_name="kunde_rechnung.html",
            pdf_filename="Rechnung Nr. "+str(self.id)+(" (Online #"+str(self.woocommerceid)+")" if self.woocommerceid else "")+".pdf",
            pdf=pdf_bestellung(self, lieferschein=False, digital=True),
            context={
                "trackinglink": str(self.trackinglink()),
                "trackingdata": bool(self.trackinglink() and self.versendet),
                "id": str(self.id),
                "woocommerceid": str(self.woocommerceid),
                "woocommercedata": bool(self.woocommerceid)
            },
            headers={"Rechnungs-ID":str(self.id)},
            bcc=[self.zahlungsempfaenger.email]
        )
        if success:
            self.rechnung_gesendet = True
            self.save()
        return success

    def html_todo_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_todonotiz_change", kwargs={"object_id": self.notiz.pk})
            return mark_safe('<a target="_blank" href="'+link+'">Notiz ansehen</a>')
        else:
            link = reverse("admin:kmuhelper_todonotiz_add")+'?from_bestellung='+str(self.pk)
            return mark_safe('<a target="_blank" href="'+link+'">Notiz hinzufügen</a>')
    html_todo_notiz.short_description = "ToDo Notiz"

    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change", kwargs={"object_id": self.notiz.pk})
            return mark_safe('<a target="_blank" href="'+link+'">Notiz ansehen</a>')
        else:
            link = reverse("admin:kmuhelper_notiz_add")+'?from_bestellung='+str(self.pk)
            return mark_safe('<a target="_blank" href="'+link+'">Notiz hinzufügen</a>')
    html_notiz.short_description = "Notiz"

    class Meta:
        verbose_name = "Bestellung"
        verbose_name_plural = "Bestellungen"


# class Gutschein(models.Model):
#     woocommerceid = models.IntegerField('WooCommerce ID', default=0)
#
#     code = models.CharField("Gutscheincode", max_length=25)
#     menge = models.FloatField("Menge (Preis oder Anzahl Prozent)")
#     typ = models.CharField("Gutscheintyp", max_length=14, choices=GUTSCHEINTYPEN)
#     beschrieb = models.TextField("Beschrieb",default="",blank=True)
#     #datum_bis = models.DateField("Gültig bis", blank=True, null=True)
#     #nicht_kumulierbar = models.BooleanField("Nicht kumulierbar", default=True, help_text="Aktivieren, damit der Gutschein nicht kumuliert werden kann.")
#
#     produkte = models.ManyToManyField("Produkt", verbose_name="Produkte", help_text="Produkte, auf welche der Gutschein angewendet werden kann.")
#     ausgeschlossene_produkte = models.ManyToManyField("Produkt", verbose_name="Ausgeschlossenes Produkt", verbose_name_plural="Ausgeschlossene Produkte",help_text="Produkte, auf welche der Gutschein nicht angewendet werden kann.")
#
#     #limit_gesamt = models.IntegerField("Gesamtlimit", default=0, help_text="Anzahl Benutzungen")
#     #limit_pro_kunde = models.IntegerField("Limit pro Kunde", default=0, help_text="Anzahl Benutzungen pro Kunde")
#     limit_artikel = models.IntegerField("Limit an Artikeln", default=0, help_text="Anzahl Artikel, auf welche der Gutschein maximal angewendet werden kann.")
#
#     #kostenlose_lieferung = models.BooleanField("Kostenlose Lieferung", default=False, help_text="Wenn aktiviert, erlaube kostenlose Lieferung")
#
#     produktkategorien = models.ManyToManyField("Kategorie", verbose_name="Kategorie", verbose_name="Kategorien",help_text="Kategorien, auf welche der Gutschein angewendet werden kann.")
#     ausgeschlossene_produktkategorien = models.ManyToManyField("Kategorie", verbose_name="Ausgeschlossene Kategorie", verbose_name_plural="Ausgeschlossene Kategorie",help_text="Kategorien, auf welche der Gutschein nicht angewendet werden kann.")
#
#     nicht_kumulierbar_mit_aktion = models.BooleanField("Nicht kumulierbar mit Aktion", default=True, help_text="Aktivieren, damit der Gutschein nicht auf Produkte angewendet werden kann, welche in Aktion sind.")
#     mindestbetrag = models.FloatField("Mindestbetrag", default=0.0)
#     maximalbetrag = models.FloatField("Maximalbetrag", default=0.0)
#
#     #erlaubte_emails
#     #benutzt_von = models.ManyToManyField("Kunde")
#
#     def gueltig_fuer_bestellung(self, bestellung):
#         if ((self.mindestbetrag <= bestellung.summe_gesamt()) and (self.maximalbetrag == 0.0 or (self.maximalbetrag >= bestellung.summe_gesamt()))):
#             return True
#         else:
#             return False
#
#     def wert_fuer_bestellung(self, bestellung):
#         if self.gueltig_fuer_bestellung(bestellung):
#             if self.typ == "fixed_cart":
#                 return self.menge
#             elif self.typ == "fixed_product":
#                 pass
#             elif self.typ == "percent":
#                 pass
#         else:
#             return 0.0
#
#     class Meta:
#         verbose_name = "Gutschein"
#         verbose_name_plural = "Gutscheine"



class Kategorie(models.Model):
    woocommerceid = models.IntegerField('WooCommerce ID', default=0)

    name = models.CharField('Name', max_length=250, default="")
    beschrieb = models.TextField('Beschrieb', default="", blank=True)
    bildlink = models.URLField('Bildlink', blank=True)

    uebergeordnete_kategorie = models.ForeignKey("self", on_delete=models.SET_NULL, null=True, blank=True, verbose_name="Übergeordnete Kategorie")

    def bild(self):
        if self.bildlink:
            return mark_safe('<img src="'+self.bildlink+'" width="100px">')
        return ""
    bild.short_description = "Bild"

    def anzahl_produkte(self):
        return self.produkte.count()
    anzahl_produkte.short_description = "Anzahl Produkte"

    def __str__(self):
        return self.name
    __str__.short_description = "Kategorie"

    class Meta:
        verbose_name = "Kategorie"
        verbose_name_plural = "Kategorien"



class Kosten(models.Model):
    name = models.CharField("Name", max_length=500, default="Zusätzliche Kosten")
    preis = models.FloatField("Preis", default=0.0)
    mwstsatz = models.FloatField('Mehrwertsteuersatz', choices= MWSTSÄTZE, default=7.7)

    def __str__(self):
        return clean(str(self.name))+(" ("+str(self.preis)+" CHF"+(" + "+str(self.mwstsatz)+"% MwSt" if self.mwstsatz else "")+")")
    __str__.short_description = "Kosten"

    class Meta:
        verbose_name = "Kosten"
        verbose_name_plural = "Kosten"



class Kunde(models.Model):
    woocommerceid = models.IntegerField('WooCommerce ID', default=0)

    email = models.EmailField("E-Mail Adresse", blank=True)
    vorname = models.CharField("Vorname", max_length=50, default="", blank=True)
    nachname = models.CharField("Nachname", max_length=50, default="", blank=True)
    firma = models.CharField("Firma", max_length=50, default="", blank=True)
    benutzername = models.CharField("Benutzername", max_length=50, default="", blank=True)
    avatar_url = models.URLField("Avatar URL", blank=True, editable=False)
    sprache = models.CharField("Sprache", default="de",choices=SPRACHEN, max_length=2)

    rechnungsadresse_vorname = models.CharField("Vorname", max_length=50, default="", blank=True)
    rechnungsadresse_nachname = models.CharField("Nachname", max_length=50, default="", blank=True)
    rechnungsadresse_firma = models.CharField("Firma", max_length=50, default="", blank=True)
    rechnungsadresse_adresszeile1 = models.CharField("Adresszeile 1", max_length=50, default="", blank=True, help_text='Strasse und Hausnummer oder "Postfach"')
    rechnungsadresse_adresszeile2 = models.CharField("Adresszeile 2", max_length=50, default="", blank=True, help_text="Wird in QR-Rechnung NICHT verwendet!")
    rechnungsadresse_ort = models.CharField("Ort", max_length=50, default="", blank=True)
    rechnungsadresse_kanton = models.CharField("Kanton", max_length=50, default="", blank=True)
    rechnungsadresse_plz = models.CharField("Postleitzahl", max_length=50, default="", blank=True)
    rechnungsadresse_land = models.CharField("Land", max_length=2, default="CH", choices=LÄNDER)
    rechnungsadresse_email = models.EmailField("E-Mail Adresse", blank=True)
    rechnungsadresse_telefon = models.CharField("Telefon", max_length=50, default="", blank=True)

    lieferadresse_vorname = models.CharField("Vorname", max_length=50, default="", blank=True)
    lieferadresse_nachname = models.CharField("Nachname", max_length=50, default="", blank=True)
    lieferadresse_firma = models.CharField("Firma", max_length=50, default="", blank=True)
    lieferadresse_adresszeile1 = models.CharField("Adresszeile 1", max_length=50, default="", blank=True)
    lieferadresse_adresszeile2 = models.CharField("Adresszeile 2", max_length=50, default="", blank=True)
    lieferadresse_ort = models.CharField("Ort", max_length=50, default="", blank=True)
    lieferadresse_kanton = models.CharField("Kanton", max_length=50, default="", blank=True)
    lieferadresse_plz = models.CharField("Postleitzahl", max_length=50, default="", blank=True)
    lieferadresse_land = models.CharField("Land", max_length=2, default="CH", choices=LÄNDER)

    zusammenfuegen = models.ForeignKey("self", on_delete=models.SET_NULL, null=True, blank=True, verbose_name="Zusammenfügen mit", help_text="Dies kann nicht widerrufen werden! Werte im aktuellen Kunden werden bevorzugt.")
    webseite = models.URLField("Webseite", blank=True, default="")
    bemerkung = models.TextField("Bemerkung", default="", blank=True)

    registrierungsemail_gesendet = models.BooleanField("Registrierungsemail gesendet?", default=False)

    def avatar(self):
        if self.avatar_url:
            return mark_safe('<img src="'+self.avatar_url+'" width="50px">')
        return ""
    avatar.short_description = "Avatar"

    def __str__(self):
        return (
            str(self.pk).zfill(8)+" " +
            (("(WC#"+str(self.woocommerceid)+") ") if self.woocommerceid else "") +
            ((self.vorname + " ") if self.vorname else "") +
            ((self.nachname + " ") if self.nachname else "") +
            ((self.firma + " ") if self.firma else "") +
            (("(" + str(self.rechnungsadresse_plz) + " " + self.rechnungsadresse_ort + ")"))
        )
    __str__.short_description = "Kunde"

    class Meta:
        verbose_name = "Kunde"
        verbose_name_plural = "Kunden"

    def save(self, *args, **kwargs):
        if self.zusammenfuegen:
            self.woocommerceid = self.woocommerceid if self.woocommerceid else self.zusammenfuegen.woocommerceid

            self.email = self.email if self.email else self.zusammenfuegen.email
            self.vorname = self.vorname if self.vorname else self.zusammenfuegen.vorname
            self.nachname = self.nachname if self.nachname else self.zusammenfuegen.nachname
            self.firma = self.firma if self.firma else self.zusammenfuegen.firma
            self.benutzername = self.benutzername if self.benutzername else self.zusammenfuegen.benutzername
            self.avatar_url = self.avatar_url if self.avatar_url else self.zusammenfuegen.avatar_url
            self.sprache = self.sprache if self.sprache != "de" else self.zusammenfuegen.sprache

            self.rechnungsadresse_vorname = self.rechnungsadresse_vorname if self.rechnungsadresse_vorname else self.zusammenfuegen.rechnungsadresse_vorname
            self.rechnungsadresse_nachname = self.rechnungsadresse_nachname if self.rechnungsadresse_nachname else self.zusammenfuegen.rechnungsadresse_nachname
            self.rechnungsadresse_firma = self.rechnungsadresse_firma if self.rechnungsadresse_firma else self.zusammenfuegen.rechnungsadresse_firma
            self.rechnungsadresse_adresszeile1 = self.rechnungsadresse_adresszeile1 if self.rechnungsadresse_adresszeile1 else self.zusammenfuegen.rechnungsadresse_adresszeile1
            self.rechnungsadresse_adresszeile2 = self.rechnungsadresse_adresszeile2 if self.rechnungsadresse_adresszeile2 else self.zusammenfuegen.rechnungsadresse_adresszeile2
            self.rechnungsadresse_ort = self.rechnungsadresse_ort if self.rechnungsadresse_ort else self.zusammenfuegen.rechnungsadresse_ort
            self.rechnungsadresse_kanton = self.rechnungsadresse_kanton if self.rechnungsadresse_kanton else self.zusammenfuegen.rechnungsadresse_kanton
            self.rechnungsadresse_plz = self.rechnungsadresse_plz if self.rechnungsadresse_plz else self.zusammenfuegen.rechnungsadresse_plz
            self.rechnungsadresse_land = self.rechnungsadresse_land if self.rechnungsadresse_land else self.zusammenfuegen.rechnungsadresse_land
            self.rechnungsadresse_email = self.rechnungsadresse_email if self.rechnungsadresse_email else self.zusammenfuegen.rechnungsadresse_email
            self.rechnungsadresse_telefon = self.rechnungsadresse_telefon if self.rechnungsadresse_telefon else self.zusammenfuegen.rechnungsadresse_telefon

            self.lieferadresse_vorname = self.lieferadresse_vorname if self.lieferadresse_vorname else self.zusammenfuegen.lieferadresse_vorname
            self.lieferadresse_nachname = self.lieferadresse_nachname if self.lieferadresse_nachname else self.zusammenfuegen.lieferadresse_nachname
            self.lieferadresse_firma = self.lieferadresse_firma if self.lieferadresse_firma else self.zusammenfuegen.lieferadresse_firma
            self.lieferadresse_adresszeile1 = self.lieferadresse_adresszeile1 if self.lieferadresse_adresszeile1 else self.zusammenfuegen.lieferadresse_adresszeile1
            self.lieferadresse_adresszeile2 = self.lieferadresse_adresszeile2 if self.lieferadresse_adresszeile2 else self.zusammenfuegen.lieferadresse_adresszeile2
            self.lieferadresse_ort = self.lieferadresse_ort if self.lieferadresse_ort else self.zusammenfuegen.lieferadresse_ort
            self.lieferadresse_kanton = self.lieferadresse_kanton if self.lieferadresse_kanton else self.zusammenfuegen.lieferadresse_kanton
            self.lieferadresse_plz = self.lieferadresse_plz if self.lieferadresse_plz else self.zusammenfuegen.lieferadresse_plz
            self.lieferadresse_land = self.lieferadresse_land if self.lieferadresse_land else self.zusammenfuegen.lieferadresse_land

            self.webseite = self.webseite if self.webseite else self.zusammenfuegen.webseite
            self.notiz = self.notiz+" "+self.zusammenfuegen.notiz

            for bestellung in self.zusammenfuegen.bestellung_set.all():
                bestellung.kunde = self
                bestellung.save()
            self.zusammenfuegen.delete()
            self.zusammenfuegen = None
        super().save(*args, **kwargs)

    def send_register_mail(self):
        self.registrierungsemail_gesendet = send_mail(
            subject="Registrierung erfolgreich!",
            to=self.email,
            template_name="kunde_registriert.html",
            context={
                "kunde": self
            },
            headers={"Kunden-ID": str(self.id)}
        )
        self.save()
        return self.registrierungsemail_gesendet

    def html_todo_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_todonotiz_change", kwargs={"object_id": self.notiz.pk})
            return mark_safe('<a target="_blank" href="'+link+'">Notiz ansehen</a>')
        else:
            link = reverse("admin:kmuhelper_todonotiz_add")+'?from_kunde='+str(self.pk)
            return mark_safe('<a target="_blank" href="'+link+'">Notiz hinzufügen</a>')
    html_todo_notiz.short_description = "ToDo Notiz"

    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change", kwargs={"object_id": self.notiz.pk})
            return mark_safe('<a target="_blank" href="'+link+'">Notiz ansehen</a>')
        else:
            link = reverse("admin:kmuhelper_notiz_add")+'?from_kunde='+str(self.pk)
            return mark_safe('<a target="_blank" href="'+link+'">Notiz hinzufügen</a>')
    html_notiz.short_description = "Notiz"



class Lieferant(models.Model):
    kuerzel = models.CharField("Kürzel", max_length=5)
    name = models.CharField('Name', max_length=50)

    webseite = models.URLField("Webseite", blank=True)
    telefon = models.CharField('Telefon', max_length=50, default="", blank=True)
    email = models.EmailField('E-Mail', null=True, blank=True)

    adresse = models.TextField('Adresse', default="", blank=True)
    notiz = models.TextField('Notiz', default="", blank=True)

    ansprechpartner = models.CharField('Ansprechpartner', max_length=250, default="", blank=True)
    ansprechpartnertel = models.CharField('Ansprechpartner Telefon', max_length=50, default="", blank=True)
    ansprechpartnermail = models.EmailField('Ansprechpartner E-Mail', null=True, default="", blank=True)

    def __str__(self):
        return self.name
    __str__.short_description = "Lieferant"

    def zuordnen(self):
        produkte = Produkt.objects.filter(lieferant=None)
        for produkt in produkte:
            produkt.lieferant = self
            produkt.save()
        return produkte.count()

    class Meta:
        verbose_name = "Lieferant"
        verbose_name_plural = "Lieferanten"



class Lieferungsposten(models.Model):
    lieferung = models.ForeignKey("Lieferung", on_delete=models.CASCADE)
    produkt = models.ForeignKey("Produkt", on_delete=models.PROTECT)
    menge = models.IntegerField("Menge", default=1)

    def __str__(self):
        return str(self.menge)+"x "+self.produkt.clean_name()
    __str__.short_description = "Lieferungsposten"

    class Meta:
        verbose_name = "Lieferungsposten"
        verbose_name_plural = "Lieferungsposten"

class Lieferung(models.Model):
    name = models.CharField("Name", max_length=50, default=defaultlieferungsname)
    datum = models.DateField("Erfasst am", auto_now_add=True)
    notiz = models.TextField("Notiz", default="", blank=True)

    lieferant = models.ForeignKey("Lieferant", on_delete=models.SET_NULL, null=True, blank=True)
    produkte = models.ManyToManyField("Produkt", through="Lieferungsposten", through_fields=("lieferung","produkt"))

    eingelagert = models.BooleanField("Eingelagert", default=False)

    def anzahlprodukte(self):
        anzahl = 0
        for i in self.produkte.through.objects.filter(lieferung=self):
            anzahl += i.menge
        return anzahl
    anzahlprodukte.short_description = "Produkte"

    def einlagern(self):
        if not self.eingelagert:
            for i in self.produkte.through.objects.filter(lieferung=self):
                i.produkt.lagerbestand += i.menge
                i.produkt.save()
            self.eingelagert = True
            self.save()
            return True
        else:
            return False

    def get_todo_notiz_link(self):
        return reverse("admin:kmuhelper_todonotiz_add")+'?from_lieferung='+str(self.id)
    get_todo_notiz_link.short_description = "ToDo Notiz"

    def __str__(self):
        return self.name
    __str__.short_description = "Lieferung"

    class Meta:
        verbose_name = "Lieferung"
        verbose_name_plural = "Lieferungen"


class Notiz(models.Model):
    name = models.CharField("Name", max_length=50)
    beschrieb = models.TextField("Beschrieb", default="", blank=True)

    erledigt = models.BooleanField("Erledigt", default=False)

    priority = models.IntegerField("Priorität", default=0, blank=True)
    erstellt_am = models.DateTimeField("Erstellt am", auto_now_add=True)

    bestellung = models.OneToOneField("Bestellung", blank=True, null=True, on_delete=models.CASCADE, related_name="notiz")
    produkt = models.OneToOneField("Produkt", blank=True, null=True, on_delete=models.CASCADE, related_name="notiz")
    kunde = models.OneToOneField("Kunde", blank=True, null=True, on_delete=models.CASCADE, related_name="notiz")

    def __str__(self):
        return self.name
    __str__.short_description = "Notiz"

    def links(self):
        text = ""
        if self.bestellung:
            text += "Bestellung <a href='"+reverse("admin:kmuhelper_bestellung_change", kwargs={"object_id":self.bestellung.pk})+"'>#"+str(self.bestellung.pk)+"</a><br>"
        if self.produkt:
            text += "Produkt <a href='"+reverse("admin:kmuhelper_produkt_change", kwargs={"object_id":self.produkt.pk})+"'>#"+str(self.produkt.pk)+"</a><br>"
        if self.kunde:
            text += "Kunde <a href='"+reverse("admin:kmuhelper_kunde_change", kwargs={"object_id":self.kunde.pk})+"'>#"+str(self.kunde.pk)+"</a><br>"
        return mark_safe(text or "Diese Notiz hat keine Verknüpfungen.")

    class Meta:
        verbose_name = "Notiz"
        verbose_name_plural = "Notizen"


class Produktkategorie(models.Model):
    produkt = models.ForeignKey("Produkt", on_delete=models.CASCADE)
    kategorie = models.ForeignKey("Kategorie", on_delete=models.CASCADE)

    def __str__(self):
        return self.kategorie.name
    __str__.short_description = "Produktkategorie"

    class Meta:
        verbose_name = "Produktkategorie"
        verbose_name_plural = "Produktkategorien"

class Produkt(models.Model):
    artikelnummer = models.CharField("Artikelnummer", max_length=25)

    woocommerceid = models.IntegerField('WooCommerce ID', default=0)

    name = models.CharField('Name', max_length=500)
    kurzbeschrieb = models.TextField('Kurzbeschrieb', default="", blank=True)
    beschrieb = models.TextField('Beschrieb', default="", blank=True)

    mengenbezeichnung = models.CharField('Mengenbezeichnung', max_length=100, default="[:de]Stück[:fr]Pièce[:it]Pezzo[:en]Piece[:]", blank=True)
    verkaufspreis = models.FloatField('Normalpreis in CHF (exkl. MwSt)', default=0)
    mwstsatz = models.FloatField('Mehrwertsteuersatz', choices= MWSTSÄTZE, default=7.7)
    lagerbestand = models.IntegerField("Lagerbestand", default=0)

    bemerkung = models.TextField('Bemerkung', default="", blank=True, help_text="Wird nicht gedruckt oder angezeigt; nur für eigene Zwecke.")
    #packlistenbemerkung = models.TextField('Packlistenbemerkung', default="", blank=True, help_text="Wird auf die Packliste gedruckt.")

    aktion_von = models.DateTimeField("In Aktion von", blank=True, null=True)
    aktion_bis = models.DateTimeField("In Aktion bis", blank=True, null=True)
    aktion_preis = models.FloatField("Aktionspreis in CHF (exkl. MwSt)", blank=True, null=True)

    datenblattlink = models.CharField('Datenblattlink', max_length=500, default="", blank=True)
    bildlink = models.URLField('Bildlink', default="", blank=True)

    lieferant = models.ForeignKey("Lieferant", on_delete=models.SET_NULL, blank=True, null=True, verbose_name="Lieferant")
    lieferant_preis = models.CharField("Lieferantenpreis", default="", blank=True, max_length=20)
    lieferant_artikelnummer = models.CharField("Lieferantenartikelnummer", default="", blank=True, max_length=25)

    kategorien = models.ManyToManyField("Kategorie", through="Produktkategorie", through_fields=("produkt","kategorie"), verbose_name="Kategorie", related_name="produkte")

    def clean_name(self, lang="de"):
        return clean(self.name,lang)
    clean_name.short_description = "Name"

    def clean_kurzbeschrieb(self, lang="de"):
        return clean(self.kurzbeschrieb,lang)
    clean_kurzbeschrieb.short_description = "Kurzbeschrieb"

    def clean_beschrieb(self, lang="de"):
        return clean(self.beschrieb,lang)
    clean_beschrieb.short_description = "Beschrieb"

    def in_aktion(self, zeitpunkt:datetime=datetime.now(utc)):
        if self.aktion_von and self.aktion_bis and self.aktion_preis:
            return bool((self.aktion_von < zeitpunkt) and (zeitpunkt < self.aktion_bis))
        return False
    in_aktion.short_description = "In Aktion"
    in_aktion.boolean = True

    def preis(self, zeitpunkt:datetime=datetime.now(utc)):
        return self.aktion_preis if self.in_aktion(zeitpunkt) else self.verkaufspreis
    preis.short_description = "Aktueller Preis in CHF (exkl. MwSt)"

    def bild(self):
        if self.bildlink:
            return mark_safe('<img src="'+self.bildlink+'" width="100px">')
        return ""
    bild.short_description = "Bild"

    def save(self, *args, **kwargs):
        if self.mengenbezeichnung == "Stück":
            self.mengenbezeichnung = "[:de]Stück[:fr]Pièce[:it]Pezzo[:en]Piece[:]"
        elif self.mengenbezeichnung == "Flasche":
            self.mengenbezeichnung = "[:de]Flasche[:fr]Bouteille[:it]Bottiglia[:en]Bottle[:]"
        elif self.mengenbezeichnung == "Tube":
            self.mengenbezeichnung = "[:de]Tube[:fr]Tube[:it]Tubetto[:en]Tube[:]"
        super().save(*args, **kwargs)

    def html_todo_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_todonotiz_change", kwargs={"object_id": self.notiz.pk})
            return mark_safe('<a target="_blank" href="'+link+'">Notiz ansehen</a>')
        else:
            link = reverse("admin:kmuhelper_todonotiz_add")+'?from_produkt='+str(self.pk)
            return mark_safe('<a target="_blank" href="'+link+'">Notiz hinzufügen</a>')
    html_todo_notiz.short_description = "ToDo Notiz"

    def html_notiz(self):
        if hasattr(self, "notiz"):
            link = reverse("admin:kmuhelper_notiz_change", kwargs={"object_id": self.notiz.pk})
            return mark_safe('<a target="_blank" href="'+link+'">Notiz ansehen</a>')
        else:
            link = reverse("admin:kmuhelper_notiz_add")+'?from_produkt='+str(self.pk)
            return mark_safe('<a target="_blank" href="'+link+'">Notiz hinzufügen</a>')
    html_notiz.short_description = "Notiz"

    def __str__(self):
        return self.artikelnummer+" - "+self.clean_name()
    __str__.short_description = "Produkt"

    class Meta:
        verbose_name = "Produkt"
        verbose_name_plural = "Produkte"



class Zahlungsempfaenger(models.Model):
    qriban = models.CharField("QR-IBAN", max_length=21+5 , validators=[RegexValidator(r'^CH[0-9]{2}\s3[0-9]{3}\s[0-9]{4}\s[0-9]{4}\s[0-9]{4}\s[0-9]{1}$', 'Bite benutze folgendes Format: CHxx 3xxx xxxx xxxx xxxx x')], help_text="QR-IBAN mit Leerzeichen")
    logourl = models.URLField("Logo (URL)", validators=[RegexValidator(r'''^[0-9a-zA-Z\-\.\|\?\(\)\*\+&"'_:;/]+\.(png|jpg)$''', '''Nur folgende Zeichen gestattet: 0-9a-zA-Z-_.:;/|?&()"'*+ - Muss auf .jpg/.png enden.''')], help_text="URL eines Bildes (.jpg/.png) - Wird auf die Rechnung gedruckt.")
    firmenname = models.CharField("Firmennname", max_length=70, help_text="Name der Firma")
    firmenuid = models.CharField("Firmen-UID", max_length=15 , validators=[RegexValidator(r'^CHE-[0-9]{3}\.[0-9]{3}\.[0-9]{3}$', 'Bite benutze folgendes Format: CHE-123.456.789')], help_text="UID der Firma - Format: CHE-123.456.789 (Mehrwertsteuernummer)")
    adresszeile1 = models.CharField("Strasse und Hausnummer oder 'Postfach'", max_length=70)
    adresszeile2 = models.CharField("PLZ und Ort", max_length=70)
    land = models.CharField("Land", max_length=2, choices=LÄNDER, default="CH")
    email = models.EmailField("E-Mail", default="", blank=True, help_text="Nicht auf der Rechnung ersichtlich")
    telefon = models.CharField("Telefon", max_length=70, default="", blank=True, help_text="Nicht auf der Rechnung ersichtlich")
    webseite = models.URLField("Webseite", help_text="Auf der Rechnung ersichtlich!")

    def __str__(self):
        return self.firmenname
    __str__.short_description = "Zahlungsempfänger"

    class Meta:
        verbose_name = "Zahlungsempfänger"
        verbose_name_plural = "Zahlungsempfänger"




######################

######################

EINSTELLUNGSTYPEN = [
    ("char", "Text"),
    ("text", "Mehrzeiliger Text"),
    ("int",  "Zahl"),
    ("floa", "Fliesskommazahl"),
    ("url",  "Url")
]

class Einstellung(models.Model):
    id   = models.CharField("ID", max_length=20, primary_key=True)
    name = models.CharField("Name", max_length=100)
    typ  = models.CharField("Typ", max_length=4, default="char", choices=EINSTELLUNGSTYPEN)

    char = models.CharField(    "Inhalt (Text)              ", max_length=250, default="",  blank=True)
    text = models.TextField(    "Inhalt (Mehrzeiliger Text) ", default="",                  blank=True)
    inte = models.IntegerField( "Inhalt (Zahl)              ", default=0,                   blank=True)
    floa = models.FloatField(   "Inhalt (Fliesskommazahl)   ", default=0.0,                 blank=True)
    url  = models.URLField(     "Inhalt (Url)               ", default="",                  blank=True)

    def __str__(self):
        return self.name
    __str__.short_description = "Einstellung"

    @property
    def inhalt(self):
        if self.typ == "char":
            return self.char
        elif self.typ == "text":
            return self.text
        elif self.typ == "int":
            return self.inte
        elif self.typ == "floa":
            return self.floa
        elif self.typ == "url":
            return self.url

    @inhalt.setter
    def inhalt(self, var):
        if self.typ == "char":
            self.char = var
        elif self.typ == "text":
            self.text = var
        elif self.typ == "int":
            self.inte = var
        elif self.typ == "floa":
            self.floa = var
        elif self.typ == "url":
            self.url = var

    def get_inhalt(self):
        return self.inhalt
    get_inhalt.short_description = "Inhalt"

    class Meta:
        verbose_name = "Einstellung"
        verbose_name_plural = "Einstellungen"



class Geheime_Einstellung(models.Model):
    id = models.CharField("ID", max_length=20, primary_key=True)
    typ  = models.CharField("Typ", max_length=4, default="char", choices=EINSTELLUNGSTYPEN)

    char = models.CharField(    "Inhalt (Text)              ", max_length=250, default="",  blank=True)
    text = models.TextField(    "Inhalt (Mehrzeiliger Text) ", default="",                  blank=True)
    inte = models.IntegerField( "Inhalt (Zahl)              ", default=0,                   blank=True)
    floa = models.FloatField(   "Inhalt (Fliesskommazahl)   ", default=0.0,                 blank=True)
    url  = models.URLField(     "Inhalt (Url)               ", default="",                  blank=True)

    def __str__(self):
        return self.id+" "+str(self.inhalt)
    __str__.short_description = "Geheime Einstellung"

    @property
    def inhalt(self):
        if self.typ == "char":
            return self.char
        elif self.typ == "text":
            return self.text
        elif self.typ == "int":
            return self.inte
        elif self.typ == "floa":
            return self.floa
        elif self.typ == "url":
            return self.url

    @inhalt.setter
    def inhalt(self, var):
        if self.typ == "char":
            self.char = var
        elif self.typ == "text":
            self.text = var
        elif self.typ == "int":
            self.inte = var
        elif self.typ == "floa":
            self.floa = var
        elif self.typ == "url":
            self.url = var

    def get_inhalt(self):
        return self.inhalt
    get_inhalt.short_description = "Inhalt"

    class Meta:
        verbose_name = "Geheime Einstellung"
        verbose_name_plural = "Geheime Einstellungen"


######################

######################

class ToDoNotizManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(erledigt=False)

class ToDoNotiz(Notiz):
    objects = ToDoNotizManager()

    class Meta:
        proxy = True
        verbose_name = "Notiz"
        verbose_name_plural = "Notizen"
        default_permissions = ('add', 'change', 'view')

class ToDoVersandManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(versendet=False)

class ToDoVersand(Bestellung):
    objects = ToDoVersandManager()

    class Meta:
        proxy = True
        verbose_name = "Bestellung"
        verbose_name_plural = "Bestellungen"
        default_permissions = ('add', 'change', 'view')

class ToDoZahlungseingangManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(bezahlt=False)

class ToDoZahlungseingang(Bestellung):
    objects = ToDoZahlungseingangManager()

    class Meta:
        proxy = True
        verbose_name = "Bestellung"
        verbose_name_plural = "Bestellung"
        default_permissions = ('add', 'change', 'view')

class ToDoLagerbestand(Produkt):
    def html_todo_notiz_erstellen(self):
        link = self.get_todo_notiz_link()
        return mark_safe('<a target="_blank" href="'+link+'">+ ToDo Notiz</a>')
    html_todo_notiz_erstellen.short_description = "ToDo Notiz Erstellen"

    def preis(self, *args, **kwargs):
        return super().preis(*args,**kwargs)
    preis.short_description = "Preis"

    def nr(self):
        return self.artikelnummer
    nr.short_description = "Nr."

    class Meta:
        proxy = True
        verbose_name = "Produkt"
        verbose_name_plural = "Produkte"
        default_permissions = ('add', 'change', 'view')

class ToDoLieferungsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(eingelagert=False)

class ToDoLieferung(Lieferung):
    def html_todo_notiz_erstellen(self):
        link = self.get_todo_notiz_link()
        return mark_safe('<a target="_blank" href="'+link+'">+ ToDo Notiz</a>')
    html_todo_notiz_erstellen.short_description = "ToDo Notiz Erstellen"

    objects = ToDoLieferungsManager()

    class Meta:
        proxy = True
        verbose_name = "Lieferung"
        verbose_name_plural = "Lieferungen"
        default_permissions = ('add', 'change', 'view')



from django.http import HttpResponse, FileResponse
from django.conf import settings
from django.core import mail
from django.template.loader import get_template
from django.utils import translation

from datetime import datetime
from io import BytesIO
import os

################

def getfirstindex(data:list, search:list):
    for s in search:
        if s in data:
            return data.index(s)
    return None

def runden(preis):
    return float("{:.2f}".format(float(round(round(preis / 0.05) * 0.05, 2))))

def formatprice(preis):
    return "{:.2f}".format(float(preis))

def clean(string, lang="de"):
    if "[:"+lang+"]" in string:
        return string.split("[:"+lang+"]")[1].split("[:")[0]
    elif "[:de]" in string:
        return string.split("[:de]")[1].split("[:")[0]
    else:
        return string

###############

def send_mail(subject:str, to:str, template_name:str, context:dict={}, headers:dict={}, bcc:list=[]):
    html_message = get_template("kmuhelper/emails/"+template_name).render(context)

    msg = mail.EmailMessage(
        subject=subject,
        body=html_message,
        to=[to],
        headers=headers,
        bcc=bcc
    )

    msg.content_subtype = "html"

    return bool(msg.send())

def send_pdf(subject:str, to:str, template_name:str, pdf:BytesIO, pdf_filename:str="file.pdf", context:dict={}, headers:dict={}, bcc:list=[]):
    html_message = get_template("kmuhelper/emails/"+template_name).render(context)

    msg = mail.EmailMessage(
        subject=subject,
        body=html_message,
        to=[to],
        headers=headers,
        bcc=bcc
    )

    msg.content_subtype = "html"
    msg.attach(filename=pdf_filename, content=pdf.read(), mimetype="application/pdf")

    return bool(msg.send())

###############

def modulo10rekursiv(strNummer):
    intTabelle = [
        [0,9,4,6,8,2,7,1,3,5],
        [9,4,6,8,2,7,1,3,5,0],
        [4,6,8,2,7,1,3,5,0,9],
        [6,8,2,7,1,3,5,0,9,4],
        [8,2,7,1,3,5,0,9,4,6],
        [2,7,1,3,5,0,9,4,6,8],
        [7,1,3,5,0,9,4,6,8,2],
        [1,3,5,0,9,4,6,8,2,7],
        [3,5,0,9,4,6,8,2,7,1],
        [5,0,9,4,6,8,2,7,1,3],
    ]
    strNummer = strNummer.replace(" ","")
    uebertrag = 0
    for num in strNummer:
        uebertrag = intTabelle[uebertrag][int(num)]
    return [0,9,8,7,6,5,4,3,2,1][uebertrag]



from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt

from urllib.parse import urlencode

from random import randint
import json

from .models import Einstellung, Geheime_Einstellung, Produkt, Kunde, Kategorie, Lieferant, Lieferung, Bestellung
from .apis import WooCommerce

# Create your views here.


@csrf_exempt
def wc_auth_key(request):
    JSON = json.loads(request.body)
    storeurl = request.headers.get("user-agent").split(";")[1].lstrip()

    url = Geheime_Einstellung.objects.get(id="wc-url")
    if url.inhalt.lstrip("https://").lstrip("http://").split("/")[0] == storeurl.lstrip("https://").lstrip("http://").split("/")[0]:
        key = Geheime_Einstellung.objects.get(id="wc-consumer_key")
        key.inhalt = JSON["consumer_key"]
        key.save()

        secret = Geheime_Einstellung.objects.get(id="wc-consumer_secret")
        secret.inhalt = JSON["consumer_secret"]
        secret.save()

        url = Einstellung.objects.get(id="wc-url")
        url.inhalt = "Bestätigt: " + storeurl + \
            " (Änderungen an diesem Eintrag werden nur nach erneutem Verbinden angewendet!)"
        url.save()
    return HttpResponse("success")


@login_required(login_url="/admin/login")
def wc_auth_end(request):
    if request.GET.get("success") == "1":
        messages.success(request, "WooCommerce erfolgreich verbunden!")
    else:
        messages.error(request, "WooCommerce konnte nicht verbunden werden!")
    return redirect(reverse('admin:kmuhelper_einstellung_changelist'))


@login_required(login_url="/admin/login")
def wc_auth_start(request):
    shopurl = Einstellung.objects.get(id="wc-url").inhalt
    if "Bestätigt" in shopurl:
        messages.error(
            request, "Bitte gib zuerst eine gültige Url ein, bevor du WooCommerce neu verbinden kannst!")
        return redirect(reverse('admin:kmuhelper_einstellung_changelist'))
    else:
        kmuhelperurl = request.get_host()
        params = {
            "app_name": "KMUHelper",
            "scope": "read",
            "user_id": randint(100000, 999999),
            "return_url": kmuhelperurl + "/kmuhelper/wc/auth/end",
            "callback_url": kmuhelperurl + "/kmuhelper/wc/auth/key"
        }
        query_string = urlencode(params)
        if not "https://" in shopurl and not "http://" in shopurl:
            shopurl = "https://" + shopurl

        url = "%s%s?%s" % (shopurl, '/wc-auth/v1/authorize', query_string)
        print(url)
        return redirect(url)


@login_required(login_url="/admin/login")
def wc_import_products(request):
    count = WooCommerce.product_import()
    messages.success(request, (('{} neue Produkte' if count !=
                                1 else 'Ein neues Produkt') + ' von WooCommerce importiert!').format(count))
    return redirect(reverse("admin:kmuhelper_produkt_changelist"))


@login_required(login_url="/admin/login")
def wc_import_customers(request):
    count = WooCommerce.customer_import()
    messages.success(request, (('{} neue Kunden' if count !=
                                1 else 'Ein neuer Kunde') + ' von WooCommerce importiert!').format(count))
    return redirect(reverse("admin:kmuhelper_kunde_changelist"))


@login_required(login_url="/admin/login")
def wc_import_categories(request):
    count = WooCommerce.category_import()
    messages.success(request, (('{} neue Kategorien' if count !=
                                1 else 'Eine neue Kategorie') + ' von WooCommerce importiert!').format(count))
    return redirect(reverse("admin:kmuhelper_kategorie_changelist"))


@login_required(login_url="/admin/login")
def wc_import_orders(request):
    count = WooCommerce.order_import()
    messages.success(request, (('{} neue Bestellungen' if count !=
                                1 else 'Eine neue Bestellung') + ' von WooCommerce importiert!').format(count))
    return redirect(reverse("admin:kmuhelper_bestellung_changelist"))


@login_required(login_url="/admin/login")
def wc_update_product(request, object_id):
    product = WooCommerce.product_update(
        Produkt.objects.get(id=int(object_id)))
    messages.success(request, "Produkt aktualisiert: " + str(product))
    return redirect(reverse("admin:kmuhelper_produkt_change", args=[object_id]))


@login_required(login_url="/admin/login")
def wc_update_customer(request, object_id):
    customer = WooCommerce.customer_update(
        Kunde.objects.get(id=int(object_id)))
    messages.success(request, "Kunde aktualisiert: " + str(customer))
    return redirect(reverse("admin:kmuhelper_kunde_change", args=[object_id]))


@login_required(login_url="/admin/login")
def wc_update_category(request, object_id):
    category = WooCommerce.category_update(
        Kategorie.objects.get(id=int(object_id)))
    messages.success(request, "Kategorie aktualisiert: " + str(category))
    return redirect(reverse("admin:kmuhelper_kategorie_change", args=[object_id]))


@login_required(login_url="/admin/login")
def wc_update_order(request, object_id):
    order = WooCommerce.order_update(Bestellung.objects.get(id=int(object_id)))
    messages.success(request, "Bestellung aktualisiert: " + str(order))
    return redirect(reverse("admin:kmuhelper_bestellung_change", args=[object_id]))


@csrf_exempt
def wc_webhooks(request):
    if "x-wc-webhook-topic" in request.headers and "x-wc-webhook-source" in request.headers:
        erwartete_url = Geheime_Einstellung.objects.get(
            id="wc-url").inhalt.lstrip("https://").lstrip("http://").split("/")[0]
        erhaltene_url = request.headers["x-wc-webhook-source"].lstrip(
            "https://").lstrip("http://").split("/")[0]
        if erhaltene_url == erwartete_url:
            print("[KMUHelper] - WooCommerce Webhook erhalten...")
            topic = request.headers["x-wc-webhook-topic"]
            obj = json.loads(request.body)
            if topic == "product.updated" or topic == "product.created":
                if Produkt.objects.filter(woocommerceid=obj["id"]).exists():
                    WooCommerce.product_update(
                        Produkt.objects.get(woocommerceid=obj["id"]), obj)
                else:
                    WooCommerce.product_create(obj)
            elif topic == "product.deleted":
                if Produkt.objects.filter(woocommerceid=obj["id"]).exists():
                    product = Produkt.objects.get(woocommerceid=obj["id"])
                    product.woocommerceid = 0
                    product.save()
            elif topic == "customer.updated" or topic == "customer.created":
                if Kunde.objects.filter(woocommerceid=obj["id"]).exists():
                    WooCommerce.customer_update(
                        Kunde.objects.get(woocommerceid=obj["id"]), obj)
                else:
                    WooCommerce.customer_create(obj)
            elif topic == "customer.deleted":
                if Kunde.objects.filter(woocommerceid=obj["id"]).exists():
                    customer = Kunde.objects.get(woocommerceid=obj["id"])
                    customer.woocommerceid = 0
                    customer.save()
            elif topic == "order.updated" or topic == "order.created":
                if Bestellung.objects.filter(woocommerceid=obj["id"]).exists():
                    WooCommerce.order_update(
                        Bestellung.objects.get(woocommerceid=obj["id"]), obj)
                else:
                    WooCommerce.order_create(obj)
            elif topic == "order.deleted":
                if Bestellung.objects.filter(woocommerceid=obj["id"]).exists():
                    order = Bestellung.objects.get(woocommerceid=obj["id"])
                    order.woocommerceid = 0
                    order.save()
        else:
            print("[KMUHelper] - WooCommerce Webhook von einer falschen Webseite ignoriert! - Erwartet: '" +
                  erwartete_url + "' - Erhalten: '" + erhaltene_url + "'")
            return HttpResponseBadRequest("Zugriff nicht gestattet!")
    return HttpResponse("Erfolgreich erhalten!")


@login_required(login_url="/admin/login")
def lieferant_zuordnen(request, object_id):
    if Lieferant.objects.filter(id=int(object_id)).exists():
        count = Lieferant.objects.get(id=int(object_id)).zuordnen()
        messages.success(request, ('Lieferant wurde ' + (('{} neuen Produkten' if count > 1 else 'einem neuen Produkt')
                                                         if count != 0 else "keinem neuen Produkt") + ' zugeordnet!').format(count))
    else:
        messages.error(request, "Lieferant konnte nicht gefunden werden!")
    return redirect(reverse("admin:kmuhelper_lieferant_change", args=[object_id]))


@login_required(login_url="/admin/login")
def lieferung_einlagern(request, object_id):
    if Lieferung.objects.filter(id=int(object_id)).exists():
        if Lieferung.objects.get(id=int(object_id)).einlagern():
            messages.success(request, "Lieferung eingelagert!")
        else:
            messages.error(
                request, "Lieferung konnte nicht eingelagert werden!")
    else:
        messages.error(request, "Lieferung konnte nicht gefunden werden!")
    return redirect(reverse("admin:kmuhelper_lieferung_change", args=[object_id]))


@login_required(login_url="/admin/login")
def kunde_email_registriert(request, object_id):
    if Kunde.objects.filter(id=int(object_id)).exists():
        if Kunde.objects.get(id=int(object_id)).send_register_mail():
            messages.success(request, "Registrierungsmail gesendet!")
        else:
            messages.error(
                request, "Registrierungsmail konnte nicht gesendet werden!")
    else:
        messages.error(request, "Kunde konnte nicht gefunden werden!")
    return redirect(reverse("admin:kmuhelper_kunde_change", args=[object_id]))


@login_required(login_url="/admin/login")
def bestellung_pdf_ansehen(request, object_id):
    if Bestellung.objects.filter(id=int(object_id)).exists():
        obj = Bestellung.objects.get(id=int(object_id))
        lieferschein = bool("lieferschein" in dict(request.GET))
        digital = not bool("druck" in dict(request.GET))
        pdf = obj.get_pdf(lieferschein=lieferschein, digital=digital)
        return pdf
    else:
        messages.error(request, "Bestellung wurde nicht gefunden!")
        return redirect(reverse("admin:kmuhelper_bestellung_changelist"))


@login_required(login_url="/admin/login")
def bestellung_pdf_an_kunden_senden(request, object_id):
    if Bestellung.objects.filter(id=int(object_id)).exists():
        if Bestellung.objects.get(id=int(object_id)).send_pdf_rechnung_to_customer():
            messages.success(request, "Rechnung an Kunden gesendet!")
        else:
            messages.error(
                request, "Rechnung konnte nicht an Kunden gesendet werden!")
    else:
        messages.error(request, "Bestellung konnte nicht gefunden werden!")
    return redirect(reverse("admin:kmuhelper_bestellung_change", args=[object_id]))


# Kunden-Entpunkte

def kunde_rechnung_ansehen(request, order_id, order_key):
    if Bestellung.objects.filter(id=int(order_id)).exists():
        obj = Bestellung.objects.get(id=int(order_id))
        lieferschein = bool("lieferschein" in dict(request.GET))
        digital = not bool("druck" in dict(request.GET))
        if obj.order_key == order_key:
            return obj.get_pdf_rechnung(lieferschein=lieferschein, digital=digital)
        else:
            return HttpResponse("Der Bestellungsschlüssel dieser Bestellung ist falsch!")
    else:
        return HttpResponse("Bestellung wurde nicht gefunden")



from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import render, redirect
from django.urls import reverse, path, reverse_lazy
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.clickjacking import xframe_options_sameorigin as allow_iframe

from .models import ToDoNotiz, ToDoVersand, ToDoZahlungseingang, ToDoLagerbestand, Lieferung

#####

@login_required(login_url=reverse_lazy("admin:login"))
def app_main(request):
    return render(request, "kmuhelper/app/main.html", {})

@allow_iframe
@login_required(login_url=reverse_lazy("admin:login"))
def app_home(request):
    return render(request, "kmuhelper/app/home.html", {
        "has_permission": True,
    })

@allow_iframe
def app_error(request):
    return render(request, "kmuhelper/app/error.html", {
        "has_permission": True,
    })

def app_manifest(request):
    response = render(request, "kmuhelper/app/manifest.webmanifest", {})
    response['Content-Type'] = 'text/json'
    response["Service-Worker-Allowed"] = "/"
    return response

#####

urlpatterns = [
    path('app/', app_main, name="app-main"),
    path('app/home', app_home, name="app-home"),
    path('app/error', app_error, name="app-error"),
    path('app/manifest.webmanifest', app_manifest, name="app-manifest"),
]
